from typing import List
from typing import Union

import numpy as np

from sonusai import SonusAIError
from sonusai.mixture.generate_truth import generate_truth
from sonusai.mixture.get_class_weights_threshold import get_class_weights_threshold
from sonusai.mixture.get_mixtures_from_mixid import get_mixtures_from_mixid


def get_class_count(mixdb: dict, record: dict, audio: np.ndarray) -> List[List[int]]:
    """Computes the number of samples for which each truth index is active for a given sample-based truth input."""
    truth_index = [sub['index'] for sub in mixdb['targets'][record['target_file_index']]['truth_settings']]

    truth = generate_truth(mixdb=mixdb,
                           record=record,
                           target=audio)

    class_weights_threshold = get_class_weights_threshold(mixdb)

    class_count = [[] for _ in range(len(truth_index))]
    for n in range(len(truth_index)):
        class_count[n] = [0] * len(truth_index[n])
        for idx, cl in enumerate(truth_index[n]):
            truth_sum = int(np.sum(truth[:, cl - 1] >= class_weights_threshold[cl - 1]))
            class_count[n][idx] = truth_sum

    return class_count


def get_total_class_count(mixdb: dict, mixid: Union[str, List[int]] = ':') -> List[int]:
    """Sums the class counts for all mixtures."""
    mixtures = get_mixtures_from_mixid(mixdb, mixid)
    return compute_total_class_count(mixdb, mixtures)


def compute_total_class_count(mixdb: dict, records: List[dict]) -> List[int]:
    total_class_count = [0] * mixdb['num_classes']
    for record in records:
        truth_indices = [sub['index'] for sub in mixdb['targets'][record['target_file_index']]['truth_settings']]
        for n in range(len(truth_indices)):
            for idx, cl in enumerate(truth_indices[n]):
                total_class_count[cl - 1] += int(record['class_count'][n][idx])

    if mixdb['truth_mutex']:
        # Compute the class count for the 'other' class
        if total_class_count[-1] != 0:
            raise SonusAIError('Error: truth_mutex was set, but the class count for the last count was non-zero.')
        total_class_count[-1] = sum([sub['samples'] for sub in records]) - sum(total_class_count)

    return total_class_count


def get_class_count_metrics_for_class(mixdb: dict, class_index: int) -> (int, int):
    """TODO"""
    pass



from copy import deepcopy
from typing import List
from typing import Union

from sonusai import SonusAIError
from sonusai.mixture.get_mixtures_from_mixid import get_mixtures_from_mixid
from sonusai.mixture.mixture_data import set_mixture_offsets


def new_mixdb_from_mixid(mixdb: dict, mixid: Union[str, List[int]]) -> dict:
    mixdb_out = deepcopy(mixdb)
    mixdb_out['mixtures'] = get_mixtures_from_mixid(mixdb_out, mixid)
    set_mixture_offsets(mixdb_out)

    if not mixdb_out['mixtures']:
        raise SonusAIError(f'Error processing mixid: {mixid}; resulted in empty list of mixtures')

    return mixdb_out



from copy import deepcopy
from typing import List
from typing import Union

import h5py
import numpy as np
from pyaaware import ForwardTransform
from pyaaware import SED

from sonusai import SonusAIError
from sonusai.mixture.get_mixtures_from_mixid import convert_mixid_to_list
from sonusai.utils import int16_to_float


def _strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def generate_truth(mixdb: dict,
                   record: dict,
                   target: np.ndarray,
                   compute: bool = True) -> np.ndarray:
    if not compute:
        return np.empty(0, dtype=np.single)

    truth = np.zeros((len(target), mixdb['num_classes']), dtype=np.single)
    for truth_setting in mixdb['targets'][record['target_file_index']]['truth_settings']:
        config = deepcopy(truth_setting)
        config['frame_size'] = mixdb['frame_size']
        config['num_classes'] = mixdb['num_classes']
        config['mutex'] = mixdb['truth_mutex']
        config['target_gain'] = record['target_gain']
        new_truth = truth_function(target=target,
                                   config=config)
        truth = truth + new_truth

    return truth


def truth_function(target: np.ndarray,
                   config: dict) -> np.ndarray:
    truth = np.zeros((len(target), config['num_classes']), dtype=np.single)

    if config['function'] == 'sed':
        if len(target) % config['frame_size'] != 0:
            raise SonusAIError(f'Number of samples in audio is not a multiple of {config["frame_size"]}')

        if 'config' not in config:
            raise SonusAIError('Truth function SED missing config')

        parameters = ['thresholds']
        for parameter in parameters:
            if parameter not in config['config']:
                raise SonusAIError(f'Truth function SED config missing required parameter: {parameter}')

        thresholds = config['config']['thresholds']
        if not _strictly_decreasing(thresholds):
            raise SonusAIError(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')

        if config['target_gain'] == 0:
            return truth

        fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        sed = SED(thresholds=thresholds,
                  index=config['index'],
                  frame_size=config['frame_size'],
                  num_classes=config['num_classes'],
                  mutex=config['mutex'])

        audio = np.int16(np.single(target) / config['target_gain'])
        for offset in range(0, len(audio), config['frame_size']):
            indices = slice(offset, offset + config['frame_size'])
            new_truth = sed.execute(fft.energy(int16_to_float(audio[indices])))
            truth[indices] = np.reshape(new_truth, (1, len(new_truth)))

        return truth

    elif config['function'] == 'file':
        if 'config' not in config:
            raise SonusAIError('Truth function file missing config')

        parameters = ['file']
        for parameter in parameters:
            if parameter not in config['config']:
                raise SonusAIError(f'Truth function file config missing required parameter: {parameter}')

        with h5py.File(name=config['config']['file'], mode='r') as f:
            truth_in = f['/truth_t'][:]

        if truth_in.ndim != 2:
            raise SonusAIError('Truth file data is not 2 dimensions')

        if truth_in.shape[0] != len(target):
            raise SonusAIError('Truth file does not contain the right amount of samples')

        if config['target_gain'] == 0:
            return truth

        if isinstance(config['index'], list):
            if len(config['index']) != truth_in.shape[1]:
                raise SonusAIError('Truth file does not contain the right amount of classes')

            truth[:, config['index']] = truth_in
        else:
            if config['index'] + truth_in.shape[1] > config['num_classes']:
                raise SonusAIError('Truth file contains too many classes')

            truth[:, config['index']:config['index'] + truth_in.shape[1]] = truth_in

        return truth

    elif config['function'] == 'energy':
        if config['target_gain'] == 0:
            return truth

        fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        for offset in range(0, len(target), config['frame_size']):
            target_energy = fft.energy(int16_to_float(target[offset:offset + config['frame_size']]))
            truth[offset:offset + config['frame_size'], config['index']] = np.single(target_energy)

        return truth

    elif config['function'] == 'phoneme':
        # Read in .txt transcript and run a Python function to generate text grid data
        # (indicating which phonemes are active)
        # Then generate truth based on this data and put in the correct classes based on config['index']
        raise SonusAIError('Truth function phoneme is not supported yet')

    raise SonusAIError(f'Unsupported truth function: {config["function"]}')


def get_truth_indices_for_mixid(mixdb: dict, mixid: Union[str, List[int]]) -> List[List[int]]:
    """Get a list of truth indices for a given mixture."""
    indices = list()

    for m in convert_mixid_to_list(mixdb, mixid):
        indices.append(get_truth_indices_for_target(mixdb, mixdb['mixtures'][m]['target_file_index']))

    return indices


def get_truth_indices_for_target(mixdb: dict, target_file_index: int) -> List[int]:
    """Get a list of truth indices for a given target index."""
    index = [sub['index'] for sub in mixdb['targets'][target_file_index]['truth_settings']]

    # flatten, uniquify, and sort
    return sorted(list(set([item for sublist in index for item in sublist])))



from typing import List


def load_mixid(mixdb: dict, name: str = None) -> List[int]:
    import json
    from os.path import exists

    from sonusai import SonusAIError

    if name is None:
        mixid = list(range(len(mixdb['mixtures'])))
    else:
        if not exists(name):
            raise SonusAIError(f'{name} does not exist')

        with open(file=name, mode='r', encoding='utf-8') as f:
            mixid = json.load(f)
            if not isinstance(mixid, dict) or 'mixid' not in mixid:
                raise SonusAIError(f'Could not find ''mixid'' in {name}')
            mixid = mixid['mixid']

    return mixid



from os.path import expandvars

import numpy as np
import sox

from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture.constants import BIT_DEPTH
from sonusai.mixture.constants import CHANNEL_COUNT
from sonusai.mixture.constants import SAMPLE_RATE


def read_audio(name: str) -> np.ndarray:
    expanded_name = expandvars(name)

    try:
        # Read in and convert to desired format
        inp = sox.Transformer()
        inp.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        return inp.build_array(input_filepath=expanded_name,
                               sample_rate_in=int(sox.file_info.sample_rate(expanded_name)))

    except Exception as e:
        if name != expanded_name:
            logger.error(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')



import numpy as np
import sox

from sonusai import SonusAIError
from sonusai.mixture.constants import BIT_DEPTH
from sonusai.mixture.constants import CHANNEL_COUNT
from sonusai.mixture.constants import SAMPLE_RATE


def apply_augmentation(audio_in: np.ndarray,
                       augmentation: dict,
                       length_common_denominator: int) -> np.ndarray:
    """Use sox to apply augmentations to audio data."""
    try:
        # Apply augmentations
        tfm = sox.Transformer()
        tfm.set_input_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        tfm.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)

        # TODO
        #  Always normalize and remove normalize from list of available augmentations
        #  Normalize to globally set level (should this be a global config parameter,
        #  or hard-coded into the script?)
        if 'normalize' in augmentation:
            tfm.norm(db_level=augmentation['normalize'])

        if 'gain' in augmentation:
            tfm.gain(gain_db=augmentation['gain'], normalize=False, limiter=True)

        if 'pitch' in augmentation:
            tfm.pitch(n_semitones=augmentation['pitch'] / 100)

        if 'tempo' in augmentation:
            factor = augmentation['tempo']
            if abs(factor - 1.0) <= 0.1:
                tfm.stretch(factor=factor)
            else:
                tfm.tempo(factor=factor, audio_type='s')

        if 'eq1' in augmentation:
            tfm.equalizer(frequency=augmentation['eq1'][0], width_q=augmentation['eq1'][1],
                          gain_db=augmentation['eq1'][2])

        if 'eq2' in augmentation:
            tfm.equalizer(frequency=augmentation['eq2'][0], width_q=augmentation['eq2'][1],
                          gain_db=augmentation['eq2'][2])

        if 'eq3' in augmentation:
            tfm.equalizer(frequency=augmentation['eq3'][0], width_q=augmentation['eq3'][1],
                          gain_db=augmentation['eq3'][2])

        if 'lpf' in augmentation:
            tfm.lowpass(frequency=augmentation['lpf'])

        # Create output data
        audio_out = tfm.build_array(input_array=audio_in, sample_rate_in=SAMPLE_RATE)

        # make sure length is multiple of length_common_denominator
        audio_len = len(audio_out)
        if audio_len % length_common_denominator:
            pad = length_common_denominator - (audio_len % length_common_denominator)
            audio_out = np.pad(audio_out, pad_width=(0, pad), mode='constant', constant_values=0)

        return audio_out
    except Exception as e:
        raise SonusAIError(f'Error applying {augmentation}: {e}')



from pyaaware import FeatureGenerator

import sonusai


def get_feature_stats(feature_mode: str,
                      frame_size: int,
                      num_classes: int,
                      truth_mutex: bool) -> (float, int, float, int, int, int):
    fg = FeatureGenerator(feature_mode=feature_mode,
                          frame_size=frame_size,
                          num_classes=num_classes,
                          truth_mutex=truth_mutex)
    num_bands = fg.num_bands
    stride = fg.stride
    step = fg.step
    decimation = fg.decimation
    transform_frame_ms = float(frame_size) / float(sonusai.mixture.SAMPLE_RATE / 1000)
    feature_ms = transform_frame_ms * decimation * stride
    feature_step_ms = transform_frame_ms * decimation * step
    feature_samples = frame_size * decimation * stride
    feature_step_samples = frame_size * decimation * step
    return feature_ms, feature_samples, feature_step_ms, feature_step_samples, num_bands, stride



import warnings
from typing import List
from typing import Union

import numpy as np
import pandas as pd

from sonusai.metrics.one_hot import one_hot
from sonusai.mixture.get_mixid_data import get_mixid_data
from sonusai.queries.queries import get_mixids_from_snr


def snr_summary(mixdb: dict,
                mixid: Union[str, List[int]],
                truth_f: np.ndarray,
                predict: np.ndarray,
                segsnr: Union[np.ndarray, None] = None,
                pred_thr: Union[float, List[float]] = 0,
                truth_thr: float = 0.5,
                timesteps: int = 0) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, dict):
    """Calculate average-over-class metrics per SNR over specified mixture list.
       Inputs:
         mixdb        mixture database (SonusAI JSON format)
         mixid
         truth_f      truth/labels      #features x num_classes
         predict      prediction data / neural net model one-hot out,  #features x num_classes
         segsnr       segmental snr from SonusAI genft,  #transform frames x 1
         pred_thr     decision threshold(s) applied to predict data, allowing predict to be
                      continuous probabilities or decisions
         truth_thr    decision threshold(s) applied to truth data, allowing truth to be
                      continuous probabilities or decisions.  Default 0.5

       Default pred_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
       if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
       the confusion matrix is calculated for all classes.

       Returns pandas dataframe (snrdf) of metrics per snr.
    """
    num_classes = truth_f.shape[1]

    mxid_snro = get_mixids_from_snr(mixdb=mixdb, mixid=mixid)

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb['truth_mutex'] and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            pred_thr = 0.5  # multi-label pred_thr scalar 0 force to 0.5 default

        if np.ndim(pred_thr) == 1:
            if len(pred_thr) == 1:
                if pred_thr[0] == 0:
                    # multi-label pred_thr array scalar 0 force to 0.5 default
                    pred_thr = 0.5
                else:
                    # multi-label pred_thr array set to scalar = array[0]
                    pred_thr = pred_thr[0]

    if segsnr is not None:
        # prep segsnr if provided, transform frames to feature frames via mean()
        # expected to always be an integer
        NFF = int(segsnr.shape[0] / truth_f.shape[0])
        segsnr_f = np.mean(np.reshape(segsnr, (truth_f.shape[0], NFF)), axis=1, keepdims=True)
        ssnr_stats = np.zeros((len(mxid_snro), 3), dtype=float)

    macro_avg = np.zeros((len(mxid_snro), 7), dtype=float)
    micro_avg = np.zeros((len(mxid_snro), 7), dtype=float)
    wghtd_avg = np.zeros((len(mxid_snro), 7), dtype=float)
    ii = 0
    for snrkey in mxid_snro:
        ytrue, ypred = get_mixid_data(mixdb, mxid_snro[snrkey], truth_f, predict)
        _, metrics, _, _, _, mavg = one_hot(ytrue, ypred, pred_thr, truth_thr, timesteps)

        # mavg macro, micro, weighted: [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
        macro_avg[ii, :] = mavg[0, 0:7]
        micro_avg[ii, :] = mavg[1, 0:7]
        wghtd_avg[ii, :] = mavg[2, 0:7]
        if segsnr is not None:
            ytrue, ysegsnr = get_mixid_data(mixdb, mxid_snro[snrkey], truth_f, segsnr_f)
            with warnings.catch_warnings():
                warnings.filterwarnings(action='ignore', message='divide by zero encountered in log10')
                # segmental SNR mean = mixture_snr and target_snr
                ssnr_stats[ii, 0] = 10 * np.log10(np.mean(ysegsnr))
                # seg SNR 80% percentile
                ssnr_stats[ii, 1] = 10 * np.log10(np.percentile(ysegsnr, 80, interpolation='midpoint'))
                # seg SNR max
                ssnr_stats[ii, 2] = 10 * np.log10(max(ysegsnr))

        ii = ii + 1

    # SNR format SNR: PPV, TPR, F1, FPR, ACC, AP, AUC, SNRSEG80P
    col_n = ['PPV', 'TPR', 'F1', 'FPR', 'ACC', 'AP', 'AUC']
    snr_macrodf = pd.DataFrame(macro_avg, index=list(mxid_snro.keys()), columns=col_n)
    snr_macrodf.sort_index(ascending=False, inplace=True)

    snr_microdf = pd.DataFrame(micro_avg, index=list(mxid_snro.keys()), columns=col_n)
    snr_microdf.sort_index(ascending=False, inplace=True)

    snr_wghtdf = pd.DataFrame(wghtd_avg, index=list(mxid_snro.keys()), columns=col_n)
    snr_wghtdf.sort_index(ascending=False, inplace=True)

    # Add segmental snr columns if provided
    if segsnr is not None:
        ssnrdf = pd.DataFrame(ssnr_stats, index=list(mxid_snro.keys()), columns=['SSNRavg', 'SSNR80p', 'SSNRmax'])
        ssnrdf.sort_index(ascending=False, inplace=True)
        snr_macrodf = pd.concat([snr_macrodf, ssnrdf], axis=1)
        snr_microdf = pd.concat([snr_microdf, ssnrdf], axis=1)
        snr_wghtdf = pd.concat([snr_wghtdf, ssnrdf], axis=1)

    return snr_macrodf, snr_microdf, snr_wghtdf, mxid_snro



from typing import List

from sonusai import SonusAIError


def get_class_weights_threshold(mixdb: dict) -> List[float]:
    """Get the class weights threshold from a mixture database."""
    class_weights_threshold = mixdb['class_weights_threshold']
    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold] * mixdb['num_classes']
    if len(class_weights_threshold) != mixdb['num_classes']:
        raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')
    return class_weights_threshold



"""sonusai vars

usage: vars

options:
    -h, --help

List custom SonusAI variables.

"""

from docopt import docopt

import sonusai
from sonusai import mixture
from sonusai.utils import trim_docstring


def main():
    docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    print('Custom SonusAI variables:\n')
    print(f'  ${{default_noise}}: {mixture.DEFAULT_NOISE}')
    print(f'  ${{frame_size}}:    {mixture.DEFAULT_FRAME_SIZE}')
    print()


if __name__ == '__main__':
    main()



import json
from typing import List
from typing import Union

import h5py
import numpy as np

from sonusai import SonusAIError
from sonusai.mixture.get_file_frame_segments import get_file_frame_segments


def get_ft_from_file(filename: str, mixid: Union[str, List[int]] = ':') -> (np.ndarray, np.ndarray):
    """Get feature/truth frames from H5 file for given mixture ID's"""
    try:
        with h5py.File(filename, 'r') as f:
            mixdb = json.loads(f.attrs['mixdb'])
            stride = f['feature'].shape[1]
            num_bands = f['feature'].shape[2]
            num_classes = f['truth_f'].shape[1]

            file_frame_segments = get_file_frame_segments(mixdb, mixid)
            total_frames = sum([file_frame_segments[m].length for m in file_frame_segments])
            feature = np.empty((total_frames, stride, num_bands), dtype=np.single)
            truth = np.empty((total_frames, num_classes), dtype=np.single)
            start = 0
            for m in file_frame_segments:
                length = file_frame_segments[m].length
                feature[start:start + length] = f['feature'][file_frame_segments[m].get_slice()]
                truth[start:start + length] = f['truth_f'][file_frame_segments[m].get_slice()]
                start += length

            return feature, truth

    except Exception as e:
        raise SonusAIError(f'Error: {e}')



import multiprocessing

import numpy


def to_numpy_array(arr: multiprocessing.Array, dtype=None) -> numpy.ndarray:
    """Convert multiprocessing array to numpy array"""
    return numpy.frombuffer(arr.get_obj(), dtype=dtype)



from typing import List
from typing import Union

import numpy as np

from sonusai.mixture import get_file_frame_segments


def get_mixid_data(mixdb: dict,
                   mixid: Union[str, List[int]],
                   truth_f: np.ndarray,
                   predict: np.ndarray) -> (np.ndarray, np.ndarray):
    """Collect per-feature data of specified mixids from mixdb where inputs are:
       truth_f:   truth data matching mixdb (size #feature_frames x num_classes)
       predict:   prediction or segsnr data size #feature_frames x ndim (ndim > 1)

    Returns:
        ytrue:    np.array combined truth from mixids
        ypred:    np.array combined data from mixids
    """
    num_classes = truth_f.shape[1]
    # same as num_class for prediction data, but use for segsnr too
    dnum = predict.shape[1]

    file_frame_segments = get_file_frame_segments(mixdb, mixid)
    total_frames = sum([file_frame_segments[m].length for m in file_frame_segments])
    ytrue = np.empty((total_frames, num_classes), dtype=np.single)
    ypred = np.empty((total_frames, dnum), dtype=np.single)

    # Handle the special case when input data is smaller, i.e., prediction when total mixture
    # length is a non-multiple of the batch size. In this case just pad both with zeros; should
    # have negligible effect on metrics
    fdiff = total_frames - truth_f.shape[0]
    if fdiff > 0:
        truth_f = np.concatenate((truth_f, np.zeros((fdiff, num_classes))))
        predict = np.concatenate((predict, np.zeros((fdiff, num_classes))))

    start = 0
    for m in file_frame_segments:
        length = file_frame_segments[m].length
        ytrue[start:start + length] = truth_f[file_frame_segments[m].get_slice()]
        ypred[start:start + length] = predict[file_frame_segments[m].get_slice()]
        start += length

    return ytrue, ypred



from sonusai.mixture.generate_truth import get_truth_indices_for_target


def get_targets_for_truth_index(mixdb: dict, truth_index: int, allow_multiple: bool = False) -> list:
    """Get a list of targets containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    target_indices = set()
    for target_index, target in enumerate(mixdb['targets']):
        indices = get_truth_indices_for_target(mixdb, target_index)
        if len(indices) == 1 or allow_multiple:
            for index in indices:
                if index == truth_index + 1:
                    target_indices.add(target_index)

    return sorted(list(target_indices))



def load_mixdb(name: str) -> dict:
    import json
    from os.path import exists
    from os.path import splitext

    import h5py

    from sonusai import SonusAIError

    if not exists(name):
        raise SonusAIError(f'{name} does not exist')

    ext = splitext(name)[1]

    if ext == '.json':
        with open(file=name, mode='r', encoding='utf-8') as f:
            return json.load(f)

    if ext == '.h5':
        with h5py.File(name=name, mode='r') as f:
            return json.loads(f.attrs['mixdb'])

    raise SonusAIError(f'Do not know how to load mixdb from {name}')



import json
import random
from copy import copy
from typing import List
from typing import Union

import h5py
import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai import SonusAIError
from sonusai.mixture import Segment
from sonusai.mixture import convert_mixid_to_list
from sonusai.mixture import get_feature_frames_in_mixture
from sonusai.mixture import get_mixtures_from_mixid
from sonusai.utils import reshape_inputs


class DataGenerator(Sequence):
    """Generates data for Keras"""

    def __init__(self,
                 filename: str,
                 mixid: Union[str, List[int]],
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        self.filename = filename
        self.mixid = mixid
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle

        try:
            self.file = h5py.File(self.filename, 'r')
            self.mixdb = json.loads(self.file.attrs['mixdb'])
            self.stride = self.file['feature'].shape[1]
            self.num_bands = self.file['feature'].shape[2]
            self.num_classes = self.file['truth_f'].shape[1]
        except Exception as e:
            raise SonusAIError(f'Error: {e}')

        self.mixid = convert_mixid_to_list(self.mixdb, self.mixid)

        self.file_frame_segments = dict()
        for m in self.mixid:
            self.file_frame_segments[m] = Segment(self.mixdb['mixtures'][m]['o_frame_offset'],
                                                  get_feature_frames_in_mixture(self.mixdb, m))

        self.mixtures = None
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None
        self.frames_per_batch = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""
        feature = np.empty((self.frames_per_batch, self.stride, self.num_bands), dtype=np.single)
        truth = np.empty((self.frames_per_batch, self.num_classes), dtype=np.single)
        start = 0
        for segment in self.batch_frame_segments[batch_index]:
            length = segment.length
            feature[start:start + length] = self.file['feature'][segment.get_slice()]
            truth[start:start + length] = self.file['truth_f'][segment.get_slice()]
            start += length

        feature, truth, _, _, _, _ = reshape_inputs(feature=feature,
                                                    truth=truth,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)
        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        if self.shuffle:
            random.shuffle(self.mixid)
            self.initialize_mixtures()

    def _initialize_mixtures(self) -> None:
        self.mixtures = get_mixtures_from_mixid(self.mixdb, self.mixid)
        self.mixture_frame_segments = [self.file_frame_segments[m] for m in self.mixid]

        frames = sum([sub['samples'] for sub in self.mixtures]) // self.mixdb['feature_step_samples']
        self.frames_per_batch = self.batch_size if self.timesteps == 0 else self.batch_size * self.timesteps
        self.total_batches = frames // self.frames_per_batch

        if self.total_batches == 0:
            raise SonusAIError(
                f'Error: dataset only contains {frames} frames which is not enough to fill a batch size of '
                f'{self.frames_per_batch}. Either provide more data or decrease the batch size')

        # Compute mixid and offset for dataset
        # offsets are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        index_map = list()
        for m in range(len(self.mixture_frame_segments)):
            current_frames = self.mixture_frame_segments[m].length
            cumulative_frames += current_frames
            while cumulative_frames >= self.frames_per_batch:
                extra_frames = cumulative_frames - self.frames_per_batch
                index_map.append({
                    'mixid':  list(range(start_mixture_index, m + 1)),
                    'offset': offset,
                    'extra':  extra_frames
                })
                if extra_frames == 0:
                    start_mixture_index = m + 1
                    offset = 0
                else:
                    start_mixture_index = m
                    offset = current_frames - extra_frames
                cumulative_frames = extra_frames

        self.batch_frame_segments = list()
        for item in index_map:
            slices = list()
            for m in item['mixid']:
                slices.append(copy(self.mixture_frame_segments[m]))
            slices[0].trim_start(item['offset'])
            slices[-1].trim_length(item['extra'])
            consolidated_segments = list()
            start = slices[0].start
            length = slices[0].length
            for i in range(1, len(slices)):
                if slices[i].start != start + length:
                    consolidated_segments.append(Segment(start, length))
                    start = slices[i].start
                    length = slices[i].length
                else:
                    length += slices[i].length
            consolidated_segments.append(Segment(start, length))

            self.batch_frame_segments.append(consolidated_segments)



from typing import List
from typing import Union


def convert_mixid_to_list(mixdb: dict, mixid: Union[str, List[int]] = None) -> List[int]:
    mixid_out = mixid

    if mixid_out is None:
        return list(range(len(mixdb['mixtures'])))

    if isinstance(mixid_out, str):
        try:
            mixid_out = list(eval(f'range(len({mixdb["mixtures"]}))[{mixid_out}]'))
        except NameError:
            return []

    if not all(isinstance(x, int) and x < len(mixdb['mixtures']) for x in mixid_out):
        return []

    return mixid_out


def get_mixtures_from_mixid(mixdb: dict, mixid: Union[str, List[int]] = None) -> list:
    mixid_out = convert_mixid_to_list(mixdb, mixid)

    return [mixdb['mixtures'][i] for i in mixid_out]



import numpy as np

from sonusai import SonusAIError


def truth_reduction(x: np.ndarray, func: str) -> np.ndarray:
    if func == 'max':
        return np.max(x, axis=0)

    if func == 'mean':
        return np.mean(x, axis=0)

    raise SonusAIError(f'Invalid truth reduction function: {func}')



"""sonusai predict

usage: predict [-hv] (-m MODEL) (-i INPUT) [-e EXPECT] [-o OUTPUT]

options:
   -h, --help
   -v, --verbose                Be verbose.
   -m MODEL, --model MODEL      Trained model ONNX file.
   -i INPUT, --input INPUT      Input feature WAV or HDF5 file.
   -e EXPECT, --expect EXPECT   Optional expected results HDF5 file.
   -o OUTPUT, --output OUTPUT   Output HDF5 file.

Run prediction on a trained model using SonusAI genft data and optionally
compare the results to expected results.

Inputs:
    MODEL   A SonusAI trained ONNX model file.
    INPUT   A WAV file; or an HDF5 file containing:
                dataset:    feature
    EXPECT  An HDF5 file containing:
                dataset:    predict

Outputs:
    OUTPUT  An HDF5 file containing:
                dataset:    predict
    predict.log

"""
import time
from os.path import exists
from os.path import splitext

import h5py
import numpy as np
from docopt import docopt
from pyaaware import FeatureGenerator
from pyaaware import ForwardTransform
from pyaaware import Predict

import sonusai
from sonusai import SonusAIError
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import mixture
from sonusai import update_console_handler
from sonusai.mixture import read_audio
from sonusai.utils import int16_to_float
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']
        model_name = args['--model']
        input_name = args['--input']
        expect_name = args['--expect']
        output_name = args['--output']

        start_time = time.monotonic()

        log_name = 'predict.log'
        create_file_handler(log_name)
        update_console_handler(verbose)
        initial_log_messages('predict')

        logger.info('')
        logger.info(f'Model:  {model_name}')
        logger.info(f'Input:  {input_name}')
        logger.info(f'Expect: {expect_name}')
        logger.info(f'Output: {output_name}')
        logger.info('')

        if not output_name and not expect_name:
            raise SonusAIError('Must specify EXPECT or OUTPUT (or both).')

        model = Predict(model_name)
        logger.debug(f'Model input shape {model.input_shape}')
        logger.debug(f'Model output shape {model.output_shape}')

        if not exists(input_name):
            raise SonusAIError(f'{input_name} does not exist')

        ext = splitext(input_name)[1]

        if ext == '.wav':
            audio = read_audio(input_name)

            fft = ForwardTransform(N=mixture.DEFAULT_FRAME_SIZE * 4, R=mixture.DEFAULT_FRAME_SIZE)
            fg = FeatureGenerator(frame_size=mixture.DEFAULT_FRAME_SIZE,
                                  feature_mode=model.feature,
                                  num_classes=1,
                                  truth_mutex=model.mutex)

            transform_frames = len(audio) // mixture.DEFAULT_FRAME_SIZE
            feature_frames = len(audio) // (mixture.DEFAULT_FRAME_SIZE * fg.decimation * fg.step)

            feature = np.empty((feature_frames, fg.stride, fg.num_bands))
            mock_truth = np.empty(fg.num_classes)

            feature_frame = 0
            for transform_frame in range(transform_frames):
                indices = slice(transform_frame * mixture.DEFAULT_FRAME_SIZE,
                                (transform_frame + 1) * mixture.DEFAULT_FRAME_SIZE)
                fd = fft.execute(int16_to_float(audio[indices]))
                fg.execute(fd, mock_truth)
                if fg.eof():
                    feature[feature_frame] = fg.feature()
                    feature_frame += 1
        elif ext == '.h5':
            with h5py.File(name=input_name, mode='r') as f:
                feature = np.array(f['feature'])
        else:
            raise SonusAIError(f'Unknown file type for {input_name}')

        logger.debug(f'Input shape {feature.shape}')

        predict = model.execute(feature)

        if expect_name:
            with h5py.File(name=expect_name, mode='r') as f:
                expected = np.array(f['predict'])
                logger.debug(f'Expect shape {expected.shape}')
                if expected.shape != predict.shape:
                    raise SonusAIError('Expect shape does not match input shape')

                max_error = np.amax(np.abs(predict - expected))
                logger.info(f'Maximum error = {max_error}')

        if output_name:
            with h5py.File(name=output_name, mode='w') as f:
                f.create_dataset(name='predict', data=predict)
                logger.info(f'Wrote {output_name}')

        end_time = time.monotonic()
        logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



import multiprocessing as mp
from typing import List
from typing import Union

import numpy as np
from tqdm import tqdm

import sonusai.mixture
from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture.augmentation_rules import get_augmentations
from sonusai.mixture.class_count import compute_total_class_count
from sonusai.mixture.get_targets_for_truth_index import get_targets_for_truth_index
from sonusai.mixture.process import process_target
from sonusai.utils import seconds_to_hms
from sonusai.utils.parallel import p_map

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


def balance_classes(mixdb: dict,
                    raw_target_audio: list,
                    logging: bool = True,
                    show_progress: bool = False) -> List[dict]:
    """Add target augmentations until the class count values are balanced."""
    first_cba_index = len(mixdb['target_augmentations'])

    MP_DICT['mixdb'] = mixdb
    MP_DICT['raw_target_audio'] = raw_target_audio

    augmented_targets = get_augmented_targets(mixdb)
    class_balancing_samples = get_class_balancing_samples(mixdb, augmented_targets)
    if logging:
        logger.info('')
        label_digits = max([len(get_class_label(mixdb, item)) for item in range(len(class_balancing_samples))])
        samples_digits = np.ceil(np.log10(float(max(class_balancing_samples))))
        samples_digits = int(samples_digits + np.ceil(samples_digits / 3))
        for class_index, required_samples in enumerate(class_balancing_samples):
            logger.info(f'Class {get_class_label(mixdb, class_index):>{label_digits}} '
                        f'needs {required_samples:>{samples_digits},} more active truth samples '
                        f' - {seconds_to_hms(required_samples / sonusai.mixture.SAMPLE_RATE)}')
        logger.info('')

    for class_index, required_samples in enumerate(class_balancing_samples):
        augmented_targets = _balance_class(mixdb=mixdb,
                                           augmented_targets=augmented_targets,
                                           first_cba_index=first_cba_index,
                                           class_index=class_index,
                                           required_samples=required_samples,
                                           logging=logging,
                                           show_progress=show_progress)

    return augmented_targets


def _balance_class(mixdb: dict,
                   augmented_targets: List[dict],
                   first_cba_index: int,
                   class_index: int,
                   required_samples: int,
                   logging: bool = True,
                   show_progress: bool = False) -> List[dict]:
    """Add target augmentations for a single class until the required samples are satisfied."""
    if required_samples == 0:
        return augmented_targets

    class_label = get_class_label(mixdb, class_index)

    # Get list of targets for this class
    target_indices = get_targets_for_truth_index(mixdb, class_index)
    if not target_indices:
        raise SonusAIError(f'Could not find single-class targets for class index {class_index}')

    num_cpus = mp.cpu_count()

    remaining_samples = required_samples
    added_samples = 0
    added_targets = 0
    progress = tqdm(total=required_samples, desc=f'Balance class {class_label}', disable=not show_progress)
    while True:
        records = []
        while len(records) < num_cpus:
            for target_index in target_indices:
                augmentation_indices = get_unused_balancing_augmentations(mixdb=mixdb,
                                                                          augmented_targets=augmented_targets,
                                                                          first_cba_index=first_cba_index,
                                                                          target_file_index=target_index,
                                                                          amount=num_cpus)
                for augmentation_index in augmentation_indices:
                    records.append({
                        'target_file_index':         target_index,
                        'target_augmentation_index': augmentation_index,
                    })

        records = records[0:num_cpus]
        records = p_map(_process_target, records)

        for record in records:
            new_samples = np.sum(np.sum(record['class_count']))
            remaining_samples -= new_samples

            # If the current record will overshoot the required samples then add it only if
            # overshooting results in a sample count closer to the required than not overshooting.
            add_record = remaining_samples >= 0 or -remaining_samples < remaining_samples + new_samples

            if add_record:
                augmented_targets.append(record)
                added_samples += new_samples
                added_targets += 1
                progress.update(new_samples)

            if remaining_samples <= 0:
                remove_unused_augmentations(mixdb=mixdb, records=augmented_targets)
                progress.update(required_samples - added_samples)
                progress.close()
                if logging:
                    logger.info(f'Added {added_targets:,} new augmented targets for class {class_label}')
                return augmented_targets


def _process_target(record: dict) -> dict:
    return process_target(record=record, mixdb=MP_DICT['mixdb'], raw_target_audio=MP_DICT['raw_target_audio'])


def get_class_balancing_samples(mixdb: dict, augmented_targets: List[dict]) -> List[int]:
    """Determine the number of additional active truth samples needed for each class in order for
    all classes to be represented evenly over all mixtures.

    If the truth mode is mutually exclusive, ignore the last class (i.e., set to zero).
    """
    class_count = compute_total_class_count(mixdb, augmented_targets)

    if mixdb['truth_mutex']:
        class_count = class_count[:-1]

    result = list(np.max(class_count) - class_count)

    if mixdb['truth_mutex']:
        result.append(0)

    return result


def get_unused_balancing_augmentations(mixdb: dict,
                                       augmented_targets: List[dict],
                                       first_cba_index: int,
                                       target_file_index: int,
                                       amount: int = 1) -> List[int]:
    """Get a list of unused balancing augmentations for a given target file index."""
    balancing_augmentations = [item for item in range(len(mixdb['target_augmentations'])) if
                               item >= first_cba_index]
    used_balancing_augmentations = [sub['target_augmentation_index'] for sub in augmented_targets if
                                    sub['target_file_index'] == target_file_index and
                                    sub['target_augmentation_index'] in balancing_augmentations]

    augmentation_indices = [item for item in balancing_augmentations if item not in used_balancing_augmentations]
    class_balancing_augmentation = get_class_balancing_augmentation(mixdb=mixdb, target_file_index=target_file_index)

    while len(augmentation_indices) < amount:
        new_augmentation = get_augmentations(class_balancing_augmentation)[0]
        mixdb['target_augmentations'].append(new_augmentation)
        augmentation_indices.append(len(mixdb['target_augmentations']) - 1)

    return augmentation_indices


def remove_unused_augmentations(mixdb: dict, records: List[dict]) -> None:
    """Remove any unused target augmentation rules from the end of the database."""
    max_used_augmentation = max([sub['target_augmentation_index'] for sub in records]) + 1
    mixdb['target_augmentations'] = mixdb['target_augmentations'][0:max_used_augmentation]


def get_class_balancing_augmentation(mixdb: dict, target_file_index: int) -> Union[dict, None]:
    """Get the class balancing augmentation rule for the given target."""
    class_balancing_augmentation = mixdb['class_balancing_augmentation']
    if 'class_balancing_augmentation' in mixdb['targets'][target_file_index]:
        class_balancing_augmentation = mixdb['targets'][target_file_index]['class_balancing_augmentation']
    return class_balancing_augmentation


def get_augmented_targets(mixdb: dict) -> List[dict]:
    """Get a list of augmented targets from a mixture database."""
    augmented_targets = list()
    keys = [
        'target_file_index',
        'target_augmentation_index',
        'samples',
        'target_gain',
        'class_count',
    ]
    snr = max(mixdb['noise_mix']['snrs'])
    mixtures = [sub for sub in mixdb['mixtures'] if sub['snr'] == snr and sub['noise_file_index'] == 0]
    for mixture in mixtures:
        record = dict()
        for key in keys:
            record[key] = mixture[key]
        augmented_targets.append(record)

    return augmented_targets


def get_class_label(mixdb: dict, class_index: int) -> str:
    if mixdb['class_labels']:
        return mixdb['class_labels'][class_index]

    return str(class_index)



import numpy as np
from sklearn.metrics import average_precision_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve

from sonusai import logger


def calculate_optimal_thresholds(truth: np.ndarray,
                                 predict: np.ndarray,
                                 truth_thr: float = 0.5) -> (np.ndarray, np.ndarray):
    """ Calculates optimal thresholds for each class from one-hot prediction and truth data
        (numpy float arrays) where both are one-hot probabilities (or quantized decisions)
        with size frames x num_classes or frames x timesteps x num_classes.

        Returns 2 threshold arrays num_classes x 1:
          thresholds_opt_pr:     optimal thesholds for PR-curve (F1) performance
          thresholds_opt_roc:    optimal thresholds for ROC-curve (TPR/FPR) performance

        Optional truth_thr is the decision threshold(s) applied to truth one-hot input thus allowing
        truth to optionally be continuous probabilities.  Default is 0.5.
    """
    if truth.shape != predict.shape:
        logger.error('Shape of truth and predict are not equal')
        exit()

    num_classes = -1
    # truth, predict can be either frames x num_classes, or frames x timesteps x num_classes
    # in binary case dim may not exist, detect this and set num_classes == 1
    if truth.ndim == 3 or (truth.ndim == 2 and num_classes == 1):
        if truth.ndim == 2:  # has timesteps = truth.shape[1]
            num_classes = 1
        else:
            num_classes = truth.shape[2]  # frames = truth.shape[0], timesteps = truth.shape[1]
        # reshape to remove timestep dimension
        truth = np.reshape(truth, (truth.shape[0] * truth.shape[1], truth.shape[2]))
        predict = np.reshape(predict, (predict.shape[0] * predict.shape[1], predict.shape[2]))
    else:
        if truth.ndim == 1:  # no timesteps dimension, = 0
            num_classes = 1
        else:
            num_classes = truth.shape[1]  # frames = truth.shape[0]

    # Apply decision to truth input
    truthb = np.int8(truth >= truth_thr)

    AP = np.zeros((num_classes, 1))
    AUC = np.zeros((num_classes, 1))
    thresholds_opt_pr = np.zeros((num_classes, 1))
    thresholds_opt_roc = np.zeros((num_classes, 1))
    eps = np.finfo(float).eps
    for nci in range(num_classes):
        # Average Precision also called area under the PR curve AUCPR and
        # AUC ROC curve using binary-ized truth and continuous prediction probabilities
        # sklearn returns nan if no active truth in a class but w/un-suppressible div-by-zero warning
        if sum(truthb[:, nci]) == 0:  # no active truth must be NaN
            thresholds_opt_pr[nci] = np.NaN
            thresholds_opt_roc[nci] = np.NaN
            AUC[nci] = np.NaN
            AP[nci] = np.NaN
        else:
            AP[nci] = average_precision_score(truthb[:, nci], predict[:, nci], average=None)
            AUC[nci] = roc_auc_score(truthb[:, nci], predict[:, nci], average=None)

            # Optimal threshold from PR curve, optimizes f-score
            precision, recall, thrpr = precision_recall_curve(truthb[:, nci], predict[:, nci])
            fscore = (2 * precision * recall) / (precision + recall + eps)
            ix = np.argmax(fscore)  # index of largest f1 score
            thresholds_opt_pr[nci] = thrpr[ix]

            # Optimal threshold from ROC curve, optimizes J-statistic (TPR-FPR) or gmean
            fpr, tpr, thrroc = roc_curve(truthb[:, nci], predict[:, nci])
            # J = tpr - fpr                  # J can result in thr > 1
            gmeans = np.sqrt(tpr * (1 - fpr))  # gmean seems better behaved
            ix = np.argmax(gmeans)
            thresholds_opt_roc[nci] = thrroc[ix]

    return thresholds_opt_pr, thresholds_opt_roc, AP, AUC



"""sonusai genmixdb

usage: genmixdb [-hv] CONFIG...

options:
   -h, --help
   -v, --verbose    Be verbose.

Create mixture database data for training and evaluation.

genmixdb creates a database of training and evaluation feature and truth data generation information. It allows
the choice of audio neural-network feature types that are supported by the Aaware real-time front-end and truth
data that is synchronized frame-by-frame with the feature data.

Here are some examples:

#### Adding target data
Suppose you have an audio file which is an example, or target, of what you want to
recognize or detect. Of course, for training a NN you also need truth data for that
file (also called labels). If you don't already have it, genmixdb can create truth using
energy-based sound detection on each frame of the feature data. You can also select
different feature types. Here's an example:

genmixdb target_gfr32ts2.yml

where target_gfr32ts2.yml contains:
---
targets:
  - data/target.wav

feature: gfr32ts2

target_augmentations:
  -
    normalize: -3.5
...

The mixture database is written to a JSON file that inherits the base name of the config file.

#### Target data mix with noise and augmentation

genmixdb mix_gfr32ts2.yml

where mix_gfr32ts2.yml contains:
---
targets:
  - data/target.wav

noise_mix:
  files:
    - data/noise.wav
  augmentations:
    - normalize: -3.5
  snrs:
    - 20

feature: gfr32ts2

output: data/my_mix.h5

target_augmentations:
  - normalize: -3.5
    pitch: [-3, 0, 3]
    tempo: [0.8, 1, 1.2]
...

In this example a time-domain mixture is created and feature data is calculated as
specified by 'feature: gfr32ts2'. Various feature types are available which vary in
spectral and temporal resolution (4 ms or higher), and other feature algorithm
parameters. The total feature size, dimension, and #frames for mixture is reported
in the log file (the log file name is derived from the output file base name; in this
case it would be mix_gfr32ts2.log).

Truth (labels) can be automatically created per feature output frame based on sound
energy detection. By default, these are appended to the feature data in a single HDF5
output file. By default, truth/label generation is turned on with default algorithm
and threshold levels (see truth section) and a single class, i.e., detecting a single
type of sound. The truth format is a single float per class representing the
probability of activity/presence, and multi-class truth/labels are possible by
specifying the number of classes and either a scalar index or a vector of indices in
which to put the truth result. For example, 'num_class: 3' and  'truth_index: 2' adds
a 1x3 vector to the feature data with truth put in index 2 (others would be 0) for
data/target.wav being an audio clip from sound type of class 2.

The mixture is created with potential data augmentation functions in the following way:
1. apply noise augmentation rule
2. apply target augmentation rule
3. adjust noise gain for specific SNR
4. add augmented noise to augmented target

The mixture length is the target length by default, and the noise signal is repeated
if it is shorter, or trimmed if longer.

#### Target and noise using path lists

Target and noise audio is specified as a list containing text files, audio files, and
file globs. Text files are processed with items on each line where each item can be a
text file, an audio file, or a file glob. Each item will be searched for audio files
which can be WAV, MP3, FLAC, AIFF, or OGG format with any sample rate, bit depth, or
channel count. All audio files will be converted to 16 kHz, 16-bit, single channel
format before processing. For example,

genmixdb dog-bark.yml

where dog-bark.yml contains:
---
targets:
  - slib/dog-outside/*.wav
  - slib/dog-inside/*.wav

will find all .wav files in the specified directories and process them as targets.

"""
import json
import time
from os.path import splitext
from random import seed

import yaml
from docopt import docopt
from tqdm import tqdm

import sonusai
from sonusai import SonusAIError
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import estimate_audio_length
from sonusai.mixture import get_augmentations
from sonusai.mixture import get_class_weights_threshold
from sonusai.mixture import get_feature_stats
from sonusai.mixture import get_input_files
from sonusai.mixture import get_max_class
from sonusai.mixture import get_total_class_count
from sonusai.mixture import load_config
from sonusai.mixture import log_duration_and_sizes
from sonusai.mixture import process_mixture
from sonusai.mixture import process_target
from sonusai.mixture import read_all_audio
from sonusai.mixture import update_truth_settings
from sonusai.utils import p_tqdm_map
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


def genmixdb(file: str = None,
             config: dict = None,
             logging: bool = True,
             show_progress: bool = False) -> dict:
    if (file is None and config is None) or (file is not None and config is not None):
        raise SonusAIError(f'Must specify either file name or config.')

    if file is not None:
        config = load_config(file)

    seed(config['seed'])

    if logging:
        logger.debug(f'Seed: {config["seed"]}')
        logger.debug('Configuration:')
        logger.debug(yaml.dump(config))

    if logging:
        logger.info('Collecting targets')
    targets = get_input_files(records=config['targets'], truth_settings=config['truth_settings'])
    if len(targets) == 0:
        raise SonusAIError('Canceled due to no targets')

    if logging:
        logger.debug('List of targets:')
        logger.debug(yaml.dump([sub['name'] for sub in targets], default_flow_style=False))

    if logging:
        logger.info('Collecting noises')
    noises = get_input_files(records=config['noise_mix']['files'])
    if logging:
        logger.debug('List of noises:')
        logger.debug(yaml.dump([sub['name'] for sub in noises], default_flow_style=False))

    if logging:
        logger.info('Collecting target augmentations')
    target_augmentations = get_augmentations(rules=config['target_augmentations'], target=True)
    expanded_target_augmentations = ''
    for augmentation in target_augmentations:
        expanded_target_augmentations += f'- {augmentation}\n'
    if logging:
        logger.debug('Expanded list of target augmentations:')
        logger.debug(expanded_target_augmentations)

    if logging:
        logger.info('Collecting noise augmentations')
    noise_augmentations = get_augmentations(config['noise_mix']['augmentations'], target=False)
    expanded_noise_augmentations = ''
    for augmentation in noise_augmentations:
        expanded_noise_augmentations += f'- {augmentation}\n'
    if logging:
        logger.debug('Expanded list of noise augmentations:')
        logger.debug(expanded_noise_augmentations)

    if logging:
        logger.debug(f'SNRs: {config["noise_mix"]["snrs"]}\n')
        logger.debug(f'Exhaustive noise: {config["noise_mix"]["exhaustive"]}\n')

    if config['truth_mode'] not in ['normal', 'mutex']:
        raise SonusAIError(f'invalid truth_mode: {config["truth_mode"]}')
    truth_mutex = config['truth_mode'] == 'mutex'
    max_class = get_max_class(config['num_classes'], truth_mutex)

    (feature_ms,
     feature_samples,
     feature_step_ms,
     feature_step_samples,
     num_bands,
     stride) = get_feature_stats(feature_mode=config['feature'],
                                 frame_size=config['frame_size'],
                                 num_classes=config['num_classes'],
                                 truth_mutex=truth_mutex)

    mixdb = {
        'class_balancing_augmentation': config['class_balancing_augmentation'],
        'class_count':                  list(),
        'class_labels':                 config['class_labels'],
        'class_weights_threshold':      get_class_weights_threshold(config),
        'feature':                      config['feature'],
        'feature_samples':              feature_samples,
        'feature_step_samples':         feature_step_samples,
        'frame_size':                   config['frame_size'],
        'mixtures':                     list(),
        'noise_mix':                    {
            'augmentations': noise_augmentations,
            'exhaustive':    config['noise_mix']['exhaustive'],
            'files':         noises,
            'snrs':          config['noise_mix']['snrs'],
        },
        'num_classes':                  config['num_classes'],
        'seed':                         config['seed'],
        'target_augmentations':         target_augmentations,
        'targets':                      targets,
        'truth_mutex':                  truth_mutex,
        'truth_reduction_function':     config['truth_reduction_function'],
        'truth_settings':               config['truth_settings'],
    }

    MP_DICT['mixdb'] = mixdb
    MP_DICT['config'] = config

    raw_noise_audio, raw_target_audio = read_all_audio(noises=noises, targets=targets, show_progress=show_progress)

    MP_DICT['raw_target_audio'] = raw_target_audio

    noise_sets = len(noises) * len(noise_augmentations)
    target_sets = len(targets) * len(target_augmentations) * len(config['noise_mix']['snrs'])
    total_mixtures = noise_sets * target_sets
    if logging:
        logger.info('')
        logger.info(f'Found {total_mixtures:,} mixtures to process')

    total_duration = 0
    for duration in [sub['duration'] for sub in targets]:
        for augmentation in target_augmentations:
            length = int(duration * sonusai.mixture.SAMPLE_RATE)
            if 'tempo' in augmentation:
                length /= augmentation['tempo']
            if length % feature_step_samples:
                length += feature_step_samples - int(length % feature_step_samples)
            total_duration += float(length) / sonusai.mixture.SAMPLE_RATE
    total_duration *= noise_sets
    total_duration *= len(config['noise_mix']['snrs'])

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=config['num_classes'],
                               feature_step_samples=feature_step_samples,
                               num_bands=num_bands,
                               stride=stride,
                               desc='Estimated')
        logger.info(f'Feature shape:        {stride} x {num_bands} ({stride * num_bands} total params)')
        logger.info(f'Feature samples:      {feature_samples} samples ({feature_ms} ms)')
        logger.info(f'Feature step samples: {feature_step_samples} samples ({feature_step_ms} ms)')

    # Get indices and offsets
    mixtures = [[] for _ in range(noise_sets)]
    n_id = 0
    for noise_index, noise in enumerate(noises):
        for noise_augmentation_index, noise_augmentation in enumerate(noise_augmentations):
            mixtures[n_id] = [[] for _ in range(target_sets)]
            t_id = 0
            noise_offset = 0
            noise_length = estimate_audio_length(audio_in=raw_noise_audio[noise_index],
                                                 augmentation=noise_augmentation,
                                                 length_common_denominator=1)
            for target_index, target in enumerate(targets):
                for target_augmentation_index, target_augmentation in enumerate(target_augmentations):
                    if not isinstance(target['truth_settings'], list):
                        target['truth_settings'] = [target['truth_settings']]

                    update_truth_settings(target['truth_settings'], config['truth_settings'])

                    for truth_setting in target['truth_settings']:
                        if not isinstance(truth_setting['index'], list):
                            truth_setting['index'] = [truth_setting['index']]

                        if any(idx > max_class for idx in truth_setting['index']):
                            raise SonusAIError('invalid truth index')

                    for snr in config['noise_mix']['snrs']:
                        mixtures[n_id][t_id] = {
                            'target_file_index':         target_index,
                            'noise_file_index':          noise_index,
                            'noise_offset':              noise_offset,
                            'target_augmentation_index': target_augmentation_index,
                            'noise_augmentation_index':  noise_augmentation_index,
                            'snr':                       snr,
                        }
                        t_id += 1

                        target_length = estimate_audio_length(audio_in=raw_target_audio[target_index],
                                                              augmentation=target_augmentation,
                                                              length_common_denominator=feature_step_samples)
                        noise_offset = int((noise_offset + target_length) % noise_length)

            n_id += 1

    # Fill in the details
    progress = tqdm(total=total_mixtures, desc='genmixdb', disable=not show_progress)
    for n_id in range(len(mixtures)):
        MP_DICT['augmented_noise_audio'] = apply_augmentation(
            audio_in=raw_noise_audio[mixtures[n_id][0]['noise_file_index']],
            augmentation=noise_augmentations[mixtures[n_id][0]['noise_augmentation_index']],
            length_common_denominator=1)
        mixtures[n_id] = p_tqdm_map(_process_mixture, mixtures[n_id], progress=progress)
    progress.close()

    # Flatten mixtures
    mixdb['mixtures'] = [item for sublist in mixtures for item in sublist]
    mixdb['class_count'] = get_total_class_count(mixdb)

    total_samples = sum([sub['samples'] for sub in mixdb['mixtures']])
    total_duration = total_samples / sonusai.mixture.SAMPLE_RATE
    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb['num_classes'],
                               feature_step_samples=feature_step_samples,
                               num_bands=num_bands,
                               stride=stride,
                               desc='Actual')

    return mixdb


def get_output_from_config(config: dict, config_name: str) -> str:
    try:
        config_base = splitext(config_name)[0]
        name = str(splitext(config['output'])[0])
        name = name.replace('${config}', config_base)
        return name
    except Exception as e:
        raise SonusAIError(f'Error getting genmixdb base name: {e}')


def _process_target(record: dict) -> dict:
    return process_target(record=record, mixdb=MP_DICT['mixdb'], raw_target_audio=MP_DICT['raw_target_audio'])


def _process_mixture(record: dict) -> dict:
    return process_mixture(record=record,
                           mixdb=MP_DICT['mixdb'],
                           raw_target_audio=MP_DICT['raw_target_audio'],
                           augmented_noise_audio=MP_DICT['augmented_noise_audio'])


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']

        for config_file in args['CONFIG']:
            start_time = time.monotonic()
            logger.info(f'Creating mixture database for {config_file}')
            config = load_config(config_file)
            output = get_output_from_config(config, config_file)

            log_name = output + '.log'
            create_file_handler(log_name)
            update_console_handler(verbose)
            initial_log_messages('genmixdb')

            mixdb = genmixdb(config=config, show_progress=True)

            json_name = output + '.json'
            with open(file=json_name, mode='w') as file:
                json.dump(mixdb, file, indent=2)
                logger.info(f'Wrote mixture database for {config_file} to {json_name}')

            end_time = time.monotonic()
            logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



from typing import List

import numpy as np

from sonusai import SonusAIError
from sonusai.mixture.apply_augmentation import apply_augmentation
from sonusai.mixture.audio_db import get_noise_audio_from_db
from sonusai.mixture.audio_db import get_target_audio_from_db
from sonusai.mixture.generate_segsnr import generate_segsnr
from sonusai.mixture.generate_truth import generate_truth
from sonusai.mixture.get_next_noise import get_next_noise


def get_target_noise_audio(mixdb: dict,
                           record: dict,
                           target_audios: List[dict],
                           noise_audios: List[dict]) -> (np.ndarray, np.ndarray):
    if record['samples'] % mixdb['frame_size'] != 0:
        raise SonusAIError(f'Number of samples in record is not a multiple of {mixdb["frame_size"]}')
    target_file_index = record['target_file_index']
    target_augmentation = mixdb['target_augmentations'][record['target_augmentation_index']]
    target_audio = apply_augmentation(audio_in=get_target_audio_from_db(target_audios, target_file_index),
                                      augmentation=target_augmentation,
                                      length_common_denominator=mixdb['feature_step_samples'])
    if len(target_audio) != record['samples']:
        raise SonusAIError('Number of samples in target does not match database')
    noise_file_index = record['noise_file_index']
    noise_augmentation_index = record['noise_augmentation_index']
    noise_audio, _ = get_next_noise(offset_in=record['noise_offset'],
                                    length=record['samples'],
                                    audio_in=get_noise_audio_from_db(noise_audios,
                                                                     noise_file_index,
                                                                     noise_augmentation_index))

    target_audio = np.array(np.single(target_audio) * record['target_snr_gain'], dtype=np.int16)
    noise_audio = np.array(np.single(noise_audio) * record['noise_snr_gain'], dtype=np.int16)

    return target_audio, noise_audio


def get_audio_and_truth_t(mixdb: dict,
                          record: dict,
                          target_audios: List[dict],
                          noise_audios: List[dict],
                          compute_truth: bool = True,
                          compute_segsnr: bool = False,
                          frame_based_segnsr: bool = False) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    target, noise = get_target_noise_audio(mixdb=mixdb,
                                           record=record,
                                           target_audios=target_audios,
                                           noise_audios=noise_audios)

    mixture = np.array(target + noise, dtype=np.int16)

    truth_t = generate_truth(mixdb=mixdb,
                             record=record,
                             target=target,
                             compute=compute_truth)

    segsnr = generate_segsnr(mixdb=mixdb,
                             record=record,
                             target=target,
                             noise=noise,
                             compute=compute_segsnr,
                             frame_based=frame_based_segnsr)

    return mixture, truth_t, target, noise, segsnr


def set_mixture_offsets(mixdb: dict,
                        initial_i_sample_offset: int = 0,
                        initial_i_frame_offset: int = 0,
                        initial_o_frame_offset: int = 0) -> None:
    i_sample_offset = initial_i_sample_offset
    i_frame_offset = initial_i_frame_offset
    o_frame_offset = initial_o_frame_offset
    for mixid in range(len(mixdb['mixtures'])):
        mixdb['mixtures'][mixid]['i_sample_offset'] = i_sample_offset
        mixdb['mixtures'][mixid]['i_frame_offset'] = i_frame_offset
        mixdb['mixtures'][mixid]['o_frame_offset'] = o_frame_offset

        i_sample_offset += get_samples_in_mixture(mixdb, mixid)
        i_frame_offset += get_transform_frames_in_mixture(mixdb, mixid)
        o_frame_offset += get_feature_frames_in_mixture(mixdb, mixid)


def get_samples_in_mixture(mixdb: dict, mixid: int) -> int:
    return mixdb['mixtures'][mixid]['samples']


def get_transform_frames_in_mixture(mixdb: dict, mixid: int) -> int:
    return mixdb['mixtures'][mixid]['samples'] // mixdb['frame_size']


def get_feature_frames_in_mixture(mixdb: dict, mixid: int) -> int:
    return mixdb['mixtures'][mixid]['samples'] // mixdb['feature_step_samples']


def get_sample_offsets_in_mixture(mixdb: dict, mixid: int) -> (int, int):
    i_sample_offset = sum([sub['samples'] for sub in mixdb['mixtures'][:mixid]])
    return i_sample_offset, i_sample_offset + mixdb['mixtures'][mixid]['samples']


def get_transform_frame_offsets_in_mixture(mixdb: dict, mixid: int) -> (int, int):
    start, stop = get_sample_offsets_in_mixture(mixdb, mixid)
    return start // mixdb['frame_size'], stop // mixdb['frame_size']


def get_feature_frame_offsets_in_mixture(mixdb: dict, mixid: int) -> (int, int):
    start, stop = get_sample_offsets_in_mixture(mixdb, mixid)
    return start // mixdb['feature_step_samples'], stop // mixdb['feature_step_samples']



import re
from copy import deepcopy
from numbers import Number
from random import uniform
from typing import Dict
from typing import List
from typing import Union

import sonusai
from sonusai import SonusAIError


def get_augmentations(rules: Union[List[Dict], Dict], target: bool = False) -> List[dict]:
    """Generate augmentations from list of input rules."""
    augmentations = list()
    if not isinstance(rules, list):
        rules = [rules]

    for in_rule in rules:
        expand_rules(augmentations, in_rule, target)

    augmentations = rand_rules(augmentations)
    return augmentations


def expand_rules(out_rules: List[dict], in_rule: dict, target: bool) -> None:
    """Expand rules."""
    # replace old 'eq' rule with new 'eq1' rule to allow both for backward compatibility
    in_rule = {'eq1' if key == 'eq' else key: value for key, value in in_rule.items()}

    for key in in_rule:
        if key not in sonusai.mixture.VALID_AUGMENTATIONS:
            raise SonusAIError(f'Invalid augmentation: {key}')

        if key in ['eq1', 'eq2', 'eq3']:
            # eq must be a list of length 3 or a list of length 3 lists
            valid = True
            multiple = False
            if isinstance(in_rule[key], list):
                if any(isinstance(el, list) for el in in_rule[key]):
                    multiple = True
                    for value in in_rule[key]:
                        if not isinstance(value, list) or len(value) != 3:
                            valid = False
                else:
                    if len(in_rule[key]) != 3:
                        valid = False
            else:
                valid = False

            if not valid:
                raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')

            if multiple:
                for value in in_rule[key]:
                    expanded_rule = deepcopy(in_rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(out_rules, expanded_rule, target)
                return

        elif key == 'count':
            pass

        else:
            if isinstance(in_rule[key], list):
                for value in in_rule[key]:
                    if isinstance(value, list):
                        raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')
                    expanded_rule = deepcopy(in_rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(out_rules, expanded_rule, target)
                return
            elif not isinstance(in_rule[key], Number):
                if not in_rule[key].startswith('rand'):
                    raise SonusAIError(f'Invalid augmentation value for {key}: {in_rule[key]}')

    out_rules.append(in_rule)


def rand_rules(in_rules: List[dict]) -> List[dict]:
    """Randomize rules."""
    out_rules = list()
    for in_rule in in_rules:
        if rule_has_rand(in_rule):
            count = 1
            if 'count' in in_rule and in_rule['count'] is not None:
                count = in_rule['count']
                del in_rule['count']
            for i in range(count):
                out_rules.append(generate_random_rule(in_rule))
        else:
            out_rules.append(in_rule)
    return out_rules


def generate_random_rule(in_rule: dict) -> dict:
    """Generate a new rule from a rule that contains 'rand' directives."""

    def rand_repl(m):
        return f'{uniform(float(m.group(1)), float(m.group(4))):.2f}'

    out_rule = deepcopy(in_rule)
    for key in out_rule:
        out_rule[key] = eval(re.sub(sonusai.mixture.RAND_PATTERN, rand_repl, str(out_rule[key])))

        # convert eq values from strings to numbers
        if key in ['eq1', 'eq2', 'eq3']:
            for n in range(3):
                if isinstance(out_rule[key][n], str):
                    out_rule[key][n] = eval(out_rule[key][n])

    return out_rule


def rule_has_rand(rule: dict) -> bool:
    """Determine if any keys in the given rule contain 'rand'?"""
    for key in rule:
        if 'rand' in str(rule[key]):
            return True

    return False



from copy import deepcopy

import yaml

import sonusai
from sonusai import SonusAIError


def raw_load_config(name: str) -> dict:
    """Load YAML file with SonusAI variable substitution."""
    with open(file=name, mode='r') as f:
        config = yaml.safe_load(f)

    config_variable_substitution(config)
    return config


def config_variable_substitution(config: dict) -> None:
    """Find custom SonusAI variables in given dictionary and substitute their values in place."""
    for key, value in config.items():
        if isinstance(value, dict):
            config_variable_substitution(value)
        else:
            if value == '${frame_size}':
                config[key] = sonusai.mixture.DEFAULT_FRAME_SIZE
            elif isinstance(value, list):
                for idx in range(len(value)):
                    if isinstance(value[idx], str):
                        value[idx] = value[idx].replace('${default_noise}', sonusai.mixture.DEFAULT_NOISE)


def get_default_config() -> dict:
    """Load default SonusAI config."""
    try:
        return raw_load_config(sonusai.mixture.DEFAULT_CONFIG)
    except Exception as e:
        raise SonusAIError(f'Error loading default config: {e}')


def load_config(name: str) -> dict:
    """Load SonusAI default config and update with given YAML file (performing SonusAI variable substitution)."""
    return update_config_from_file(name=name, config=get_default_config())


def update_config_from_file(name: str, config: dict) -> dict:
    """Update the given config with the config in the given YAML file."""
    new_config = deepcopy(config)

    try:
        given_config = raw_load_config(name)
    except Exception as e:
        raise SonusAIError(f'Error loading config from {name}: {e}')

    # Use default config as base and overwrite with given config keys as found
    for key in new_config:
        if key in given_config:
            if key not in ['noise_mix', 'truth_settings']:
                new_config[key] = given_config[key]

    # Handle 'noise_mix' special case
    if 'noise_mix' in given_config:
        new_config['noise_mix'] = deepcopy(given_config['noise_mix'])
    default = deepcopy(config['noise_mix'])
    update_noise_mix(new_config['noise_mix'], default)

    # Handle 'truth_settings' special case
    if 'truth_settings' in given_config:
        new_config['truth_settings'] = deepcopy(given_config['truth_settings'])

    if not isinstance(new_config['truth_settings'], list):
        new_config['truth_settings'] = [new_config['truth_settings']]

    default = deepcopy(config['truth_settings'])
    if not isinstance(default, list):
        default = [default]

    update_truth_settings(new_config['truth_settings'], default)

    # Check for required keys
    required_keys = [
        'class_labels',
        'class_weights_threshold',
        'feature',
        'frame_size',
        'noise_mix',
        'num_classes',
        'seed',
        'targets',
        'target_augmentations',
        'class_balancing_augmentation',
        'truth_settings',
        'truth_mode',
        'truth_reduction_function',
    ]
    for key in required_keys:
        if key not in new_config:
            raise SonusAIError(f"Missing required '{key}' in {name}")

    return new_config


def update_noise_mix(given: dict, default: dict = None) -> None:
    """Update missing fields in given 'noise_mix' with default values."""
    required_keys = [
        'files',
        'augmentations',
        'snrs',
        'exhaustive',
    ]
    for key in required_keys:
        if key not in given:
            if default is not None and key in default:
                given[key] = default[key]
            else:
                raise SonusAIError(f"Missing required '{key}' in noise_mix")

    if not given['snrs']:
        given['snrs'] = [99]


def update_truth_settings(given: list, default: list = None) -> None:
    """Update missing fields in given 'truth_settings' with default values."""
    if default is not None and len(given) != len(default):
        raise SonusAIError(f'Length of truth_settings does not match given default')

    required_keys = [
        'function',
        'config',
        'index',
    ]
    for n in range(len(given)):
        for key in required_keys:
            if key not in given[n]:
                if default is not None and key in default[n]:
                    given[n][key] = default[n][key]
                else:
                    raise SonusAIError(f"Missing required '{key}' in truth_settings")


def get_hierarchical_config_files(root: str, leaf: str) -> list:
    """Get a hierarchical list of config files in the given leaf of the given root."""
    import os
    from pathlib import Path

    config_file = 'config.yml'

    root_path = Path(os.path.abspath(root))
    if not root_path.is_dir():
        raise SonusAIError(f'Given root, {root_path}, is not a directory.')

    leaf_path = Path(os.path.abspath(leaf))
    if not leaf_path.is_dir():
        raise SonusAIError(f'Given leaf, {leaf_path}, is not a directory.')

    common = os.path.commonpath((root_path, leaf_path))
    if os.path.normpath(common) != os.path.normpath(root_path):
        raise SonusAIError(f'Given leaf, {leaf_path}, is not in the hierarchy of the given root, {root_path}')

    top_config_file = Path(os.path.join(root_path, config_file))
    if not top_config_file.is_file():
        raise SonusAIError(f'Could not find {top_config_file}')

    current = leaf_path
    config_files = list()
    while current != root_path:
        local_config_file = Path(os.path.join(current, config_file))
        if local_config_file.is_file():
            config_files.append(local_config_file)
        current = current.parent

    config_files.append(top_config_file)
    return list(reversed(config_files))


def update_config_from_hierarchy(root: str, leaf: str, config: dict) -> dict:
    """Update the given config using the hierarchical config files in the given leaf of the given root."""
    new_config = deepcopy(config)
    config_files = get_hierarchical_config_files(root=root, leaf=leaf)
    for config_file in config_files:
        new_config = update_config_from_file(name=config_file, config=new_config)

    return new_config


def get_max_class(num_classes: int, truth_mutex: bool) -> int:
    """Get the maximum class index."""
    max_class = num_classes
    if truth_mutex:
        max_class -= 1
    return max_class



from typing import List
from typing import Union

import numpy as np
import pandas as pd

from sonusai.metrics.one_hot import one_hot
from sonusai.mixture.get_mixid_data import get_mixid_data


def confusion_matrix_summary(mixdb: dict,
                             mixid: Union[str, List[int]],
                             truth_f: np.ndarray,
                             predict: np.ndarray,
                             class_idx: int,
                             pred_thr: Union[float, List[float]] = 0,
                             truth_thr: float = 0.5,
                             timesteps: int = 0) -> (pd.DataFrame, pd.DataFrame):
    """Calculate confusion matrix for specified class, using truth and prediction
       data features x num_classes.

       pred_thr sets the decision threshold(s) applied to predict data, thus allowing
       predict to be continuous probabilities.

       Default pred_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
       if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
       the confusion matrix is calculated for all classes.

       Returns pandas dataframes of confusion matrix cmdf and normalized confusion matrix cmndf.
    """
    num_classes = truth_f.shape[1]
    ytrue, ypred = get_mixid_data(mixdb=mixdb, mixid=mixid, truth_f=truth_f, predict=predict)

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb['truth_mutex'] and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            # multi-label pred_thr scalar 0 force to 0.5 default
            pred_thr = 0.5

        if np.ndim(pred_thr) == 1:
            if len(pred_thr) == 1:
                if pred_thr[0] == 0:
                    # multi-label pred_thr array scalar 0 force to 0.5 default
                    pred_thr = 0.5
                else:
                    # multi-label pred_thr array set to scalar = array[0]
                    pred_thr = pred_thr[0]
            else:
                # multi-label pred_thr array scalar set = array[class_idx]
                pred_thr = pred_thr[class_idx]

    if len(mixdb['class_labels']) == num_classes:
        class_names = mixdb['class_labels']
    else:
        class_names = ([f'Class {i}' for i in range(1, num_classes + 1)])

    class_nums = ([f'{i}' for i in range(1, num_classes + 1)])

    if mixdb['truth_mutex']:
        # single-label mode force to argmax mode
        pred_thr = 0
        _, _, cm, cmn, _, _ = one_hot(ytrue, ypred, pred_thr, truth_thr, timesteps)
        row_n = class_names
        row_n[-1] = 'Other'
        # mux = pd.MultiIndex.from_product([['Single-label/mutex mode, truth thr = {}'.format(truth_thr)],
        #                                   class_nums])
        # mux = pd.MultiIndex.from_product([['truth thr = {}'.format(truth_thr)], class_nums])

        cmdf = pd.DataFrame(cm, index=row_n, columns=class_nums, dtype=int)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=class_nums, dtype=float)

    else:
        _, _, cm, cmn, _, _ = one_hot(ytrue[:, class_idx], ypred[:, class_idx], pred_thr, truth_thr, timesteps)
        cname = class_names[class_idx]
        row_n = ['TrueN', 'TrueP']
        col_n = ['N-' + cname, 'P-' + cname]
        cmdf = pd.DataFrame(cm, index=row_n, columns=col_n, dtype=int)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=col_n, dtype=float)
        # add thresholds in 3rd row
        pdnote = pd.DataFrame(np.atleast_2d([pred_thr, truth_thr]), index=['p/t thr:'], columns=col_n)
        cmdf = pd.concat([cmdf, pdnote])
        cmndf = pd.concat([cmndf, pdnote])

    return cmdf, cmndf



"""sonusai evaluate

usage: evaluate [-hv] (-f FEATURE) (-n PREDICT) [-t PTHR] [-p PLOTNUM]

options:
   -h, --help
   -v, --verbose                    Be verbose.
   -f FEATURE, --feature FEATURE    Feature + truth .h5 data file
   -n PREDICT, --predict PREDICT    Predict .h5 data file.
   -t PTHR, --thr PTHR              Optional prediction decision threshold, 0 = argmax(). [default: 0].
   -p PLOTNUM, --plotnum PLOTNUM    Optional plot mixture results (-1 is plot all, 0 is plot none) [default: 0].

Evaluate calculates performance metrics of neural-network models from model prediction data and the associated
feature+truth data file.

Inputs:
    FEATURE     A SonusAI feature+truth HDF5 file. Contains:
                    attribute:  mixdb
                    dataset:    feature   Note: not required for metrics
                    dataset:    truth_f
                    dataset:    segsnr    Optional
    PREDICT     A SonusAI predict HDF5 file. Contains:
                    dataset:    predict (either [frames, num_classes] or [frames, timesteps, num_classes])
"""
import json
from datetime import datetime
from os import mkdir
from os.path import join
from pathlib import Path
from typing import Union

import h5py
import numpy as np
from docopt import docopt

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.metrics import calculate_optimal_thresholds
from sonusai.metrics import class_summary
from sonusai.metrics import snr_summary
from sonusai.queries.queries import get_mixids_from_snr
from sonusai.utils import human_readable_size
from sonusai.utils import read_predict_data
from sonusai.utils import reshape_truth_predict
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring


def evaluate(mixdb: dict,
             truth: np.ndarray,
             predict: Union[None, np.ndarray],
             segsnr: Union[None, np.ndarray] = None,
             output_dir: Union[None, str, Path] = None,
             pred_thr: Union[float, np.ndarray] = 0.0,
             feature: Union[None, np.ndarray] = None,
             verbose: bool = False) -> None:
    update_console_handler(verbose)
    initial_log_messages('evaluate')

    if truth.shape[-1] != predict.shape[-1]:
        logger.exception(f'Number of classes in truth and predict are not equal. Exiting.')
        raise SystemExit(1)

    # truth, predict can be either frames x num_classes, or frames x timesteps x num_classes
    # in binary case dim may not exist, detect this and set num_classes == 1
    timesteps = -1
    truth, predict, num_classes = reshape_truth_predict(truth, predict, timesteps)

    fdiff = truth.shape[0] - predict.shape[0]
    if fdiff > 0:
        # truth = truth[0:-fdiff,:]
        predict = np.concatenate((predict, np.zeros((fdiff, num_classes))))
        logger.info(f'Truth has more feature-frames than predict, padding predict with zeros to match.')

    if fdiff < 0:
        predict = predict[0:fdiff, :]
        logger.info(f'Predict has more feature-frames than truth, trimming predict to match.')

    # Check segsnr, input is always in transform frames
    compute_segsnr = False
    if len(segsnr) > 0:
        segsnr_feature_frames = segsnr.shape[0] / (mixdb['feature_step_samples'] / mixdb['frame_size'])
        if segsnr_feature_frames == truth.shape[0]:
            compute_segsnr = True
        else:
            logger.warning('segsnr length does not match truth, ignoring.')

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb['truth_mutex']:
        if num_classes > 1:
            if np.ndim(pred_thr) == 0 and pred_thr == 0:
                pred_thr = 0.5  # multi-label pred_thr scalar 0 force to 0.5 default

            if np.ndim(pred_thr) == 1:
                if len(pred_thr) == 1:
                    if pred_thr[0] == 0:
                        pred_thr = 0.5  # multi-label pred_thr array scalar 0 force to 0.5 default
                    else:
                        pred_thr = pred_thr[0]  # multi-label pred_thr array set to scalar = array[0]

    else:
        pred_thr = 0  # single-label mode, force argmax mode

    if pred_thr == -1:
        thrpr, thrroc, _, _ = calculate_optimal_thresholds(truth, predict)
        pred_thr = np.atleast_1d(thrpr)
        pred_thr = np.maximum(pred_thr, 0.1)  # enforce lower limit 0.1
        pred_thr = np.minimum(pred_thr, 0.9)  # enforce upper limit 0.9
        pred_thr = pred_thr.round(2)

    # Summarize the mixture data
    num_mixtures = len(mixdb['mixtures'])
    total_samples = sum([sub['samples'] for sub in mixdb['mixtures']])
    duration = total_samples / sonusai.mixture.SAMPLE_RATE

    logger.info('')
    logger.info(f'Mixtures: {num_mixtures}')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'truth:    {human_readable_size(truth.nbytes, 1)}')
    logger.info(f'predict:  {human_readable_size(predict.nbytes, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')
    if feature:
        logger.info(f'feature:  {human_readable_size(feature.nbytes, 1)}')

    logger.info(f'Classes: {num_classes}')
    if mixdb['truth_mutex']:
        logger.info(f'Mode:  Single-label / truth_mutex / softmax')
    else:
        logger.info(f'Mode:  Multi-label / Binary')

    mxid_snro = get_mixids_from_snr(mixdb=mixdb)
    snrlist = list(mxid_snro.keys())
    snrlist.sort(reverse=True)
    logger.info(f'Ordered SNRs: {snrlist}\n')
    if type(pred_thr) is np.ndarray:
        logger.info(f'Prediction Threshold(s): {pred_thr.transpose()}\n')
    else:
        logger.info(f'Prediction Threshold(s): {pred_thr}\n')

    # Top-level report over all mixtures
    if compute_segsnr:
        macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth, predict, segsnr, pred_thr)
    else:
        macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth, predict, [], pred_thr)

    if num_classes > 1:
        logger.info(f'Metrics micro-avg per SNR over all {num_mixtures} mixtures:')
    else:
        logger.info(f'Metrics per SNR over all {num_mixtures} mixtures:')
    logger.info(microdf.round(3).to_string())
    logger.info('\n')
    if output_dir:
        microdf.round(3).to_csv(join(output_dir, 'snr.csv'))

    if mixdb['truth_mutex']:
        if compute_segsnr:
            macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth[:, 0:-1], predict[:, 0:-1], segsnr,
                                                              pred_thr)
        else:
            macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth[:, 0:-1], predict[:, 0:-1], [],
                                                              pred_thr)

        logger.info(f'Metrics micro-avg without "Other" class per SNR over all {num_mixtures} mixtures:')
        logger.info(microdf.round(3).to_string())
        logger.info('\n')
        if output_dir:
            microdf.round(3).to_csv(join(output_dir, 'snrwo.csv'))

    for snri in snrlist:
        mxids = mxid_snro[snri]
        classdf = class_summary(mixdb, mxids, truth, predict, pred_thr)
        logger.info(f'Metrics per class for SNR {snri} over {len(mxids)} mixtures:')
        logger.info(classdf.round(3).to_string())
        logger.info('\n')
        if output_dir:
            classdf.round(3).to_csv(join(output_dir, f'class_snr{snri}.csv'))


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        feature_name = args['--feature']
        predict_name = args['--predict']
        predict_threshold = float(args['--thr'])
        plot_number = int(args['--plotnum'])

        # create output directory
        output_dir = f'evaluate-{datetime.now():%Y%m%d}'
        try:
            mkdir(output_dir)
        except OSError as error:
            output_dir = f'evaluate-{datetime.now():%Y%m%d-%H%M%S}'
            try:
                mkdir(output_dir)
            except OSError as error:
                logger.error(f'Could not create directory, {output_dir}: {error}')
                raise SystemExit(1)

        log_name = output_dir + '/evaluate.log'
        create_file_handler(log_name)

        # mixdb, feature, truth, segsnr = read_feature_data(feature_name)
        with h5py.File(feature_name, 'r') as f:
            mixdb = json.loads(f.attrs['mixdb'])
            # feature = f['/feature'][:]  # No feature load saves significant time
            truth_f = f['/truth_f'][:]
            segsnr = f['/segsnr'][:]

        # mixture, target, noise = read_mixture_data(mixture_name)
        predict = read_predict_data(predict_name)

        evaluate(mixdb=mixdb,
                 truth=truth_f,
                 segsnr=segsnr,
                 output_dir=output_dir,
                 predict=predict,
                 pred_thr=predict_threshold,
                 verbose=args['--verbose'])

        logger.info(f'Wrote results to {output_dir}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



# SonusAI mixture utilities
from sonusai.mixture.apply_augmentation import apply_augmentation
from sonusai.mixture.audio_db import build_noise_audio_db
from sonusai.mixture.audio_db import build_target_audio_db
from sonusai.mixture.audio_db import check_audio_files_exist
from sonusai.mixture.audio_db import get_noise_audio_from_db
from sonusai.mixture.audio_db import get_target_audio_from_db
from sonusai.mixture.audio_db import read_all_audio
from sonusai.mixture.augmentation_rules import get_augmentations
from sonusai.mixture.class_balancing import balance_classes
from sonusai.mixture.class_balancing import get_augmented_targets
from sonusai.mixture.class_balancing import get_class_balancing_samples
from sonusai.mixture.class_balancing import get_unused_balancing_augmentations
from sonusai.mixture.class_count import compute_total_class_count
from sonusai.mixture.class_count import get_class_count
from sonusai.mixture.class_count import get_total_class_count
from sonusai.mixture.config import get_default_config
from sonusai.mixture.config import get_max_class
from sonusai.mixture.config import load_config
from sonusai.mixture.config import raw_load_config
from sonusai.mixture.config import update_config_from_file
from sonusai.mixture.config import update_config_from_hierarchy
from sonusai.mixture.config import update_truth_settings
from sonusai.mixture.constants import BIT_DEPTH
from sonusai.mixture.constants import CHANNEL_COUNT
from sonusai.mixture.constants import DEFAULT_CONFIG
from sonusai.mixture.constants import DEFAULT_FRAME_SIZE
from sonusai.mixture.constants import DEFAULT_NOISE
from sonusai.mixture.constants import FLOAT_BYTES
from sonusai.mixture.constants import RAND_PATTERN
from sonusai.mixture.constants import SAMPLE_BYTES
from sonusai.mixture.constants import SAMPLE_RATE
from sonusai.mixture.constants import VALID_AUGMENTATIONS
from sonusai.mixture.estimate_audio_length import estimate_audio_length
from sonusai.mixture.generate_segsnr import generate_segsnr
from sonusai.mixture.generate_truth import generate_truth
from sonusai.mixture.generate_truth import get_truth_indices_for_mixid
from sonusai.mixture.generate_truth import get_truth_indices_for_target
from sonusai.mixture.generate_truth import truth_function
from sonusai.mixture.get_class_weights_threshold import get_class_weights_threshold
from sonusai.mixture.get_feature_stats import get_feature_stats
from sonusai.mixture.get_file_frame_segments import get_file_frame_segments
from sonusai.mixture.get_ft_from_file import get_ft_from_file
from sonusai.mixture.get_mixid_data import get_mixid_data
from sonusai.mixture.get_mixtures_from_mixid import convert_mixid_to_list
from sonusai.mixture.get_mixtures_from_mixid import get_mixtures_from_mixid
from sonusai.mixture.get_next_noise import get_next_noise
from sonusai.mixture.get_offsets import get_offsets
from sonusai.mixture.get_targets_for_truth_index import get_targets_for_truth_index
from sonusai.mixture.input_files import get_input_files
from sonusai.mixture.load_mixdb import load_mixdb
from sonusai.mixture.load_mixid import load_mixid
from sonusai.mixture.log_duration_and_sizes import log_duration_and_sizes
from sonusai.mixture.mixture_data import get_audio_and_truth_t
from sonusai.mixture.mixture_data import get_feature_frame_offsets_in_mixture
from sonusai.mixture.mixture_data import get_feature_frames_in_mixture
from sonusai.mixture.mixture_data import get_sample_offsets_in_mixture
from sonusai.mixture.mixture_data import get_samples_in_mixture
from sonusai.mixture.mixture_data import get_target_noise_audio
from sonusai.mixture.mixture_data import get_transform_frame_offsets_in_mixture
from sonusai.mixture.mixture_data import get_transform_frames_in_mixture
from sonusai.mixture.mixture_data import set_mixture_offsets
from sonusai.mixture.new_mixdb_from_mixid import new_mixdb_from_mixid
from sonusai.mixture.process import process_mixture
from sonusai.mixture.process import process_target
from sonusai.mixture.read_audio import read_audio
from sonusai.mixture.segment import Segment
from sonusai.mixture.truth_reduction import truth_reduction



import os
import re

import sonusai

VALID_AUGMENTATIONS = ['normalize', 'gain', 'pitch', 'tempo', 'eq1', 'eq2', 'eq3', 'lpf', 'count']
RAND_PATTERN = re.compile(r'rand\(([-+]?([0-9]+(\.[0-9]*)?|\.[0-9]+)),\s*([-+]?([0-9]+(\.[0-9]*)?|\.[0-9]+))\)')
SAMPLE_RATE = 16000
BIT_DEPTH = 16
CHANNEL_COUNT = 1
SAMPLE_BYTES = int(BIT_DEPTH / 8)
FLOAT_BYTES = 4
DEFAULT_FRAME_SIZE = 64

DEFAULT_NOISE = os.path.join(sonusai.BASEDIR, 'data', 'whitenoise.wav')
DEFAULT_CONFIG = os.path.join(sonusai.BASEDIR, 'data', 'genmixdb.yml')



from typing import List

import numpy as np

from sonusai.mixture.apply_augmentation import apply_augmentation
from sonusai.mixture.class_count import get_class_count
from sonusai.mixture.get_next_noise import get_next_noise


def _process_target_audio(mixdb: dict, record: dict, raw_target_audio: List[np.ndarray]) -> np.ndarray:
    """Apply augmentation and update target metadata."""
    target_augmentation = mixdb['target_augmentations'][record['target_augmentation_index']]

    augmented_target_audio = apply_augmentation(audio_in=raw_target_audio[record['target_file_index']],
                                                augmentation=target_augmentation,
                                                length_common_denominator=mixdb['feature_step_samples'])

    record['samples'] = len(augmented_target_audio)

    # target_gain is used to back out the gain augmentation in order to return the target audio
    # to its normalized level when calculating truth.
    if 'gain' in target_augmentation:
        record['target_gain'] = 10 ** (target_augmentation['gain'] / 20)
    else:
        record['target_gain'] = 1
    record['class_count'] = get_class_count(mixdb=mixdb, record=record, audio=augmented_target_audio)

    return augmented_target_audio


def process_target(record: dict, mixdb: dict, raw_target_audio: List[np.ndarray]) -> dict:
    """Apply augmentation and update target metadata."""
    _process_target_audio(mixdb, record, raw_target_audio)
    return record


def process_mixture(record: dict,
                    mixdb: dict,
                    raw_target_audio: List[np.ndarray],
                    augmented_noise_audio: np.ndarray) -> dict:
    augmented_target_audio = _process_target_audio(mixdb, record, raw_target_audio)

    noise_segment, _ = get_next_noise(offset_in=record['noise_offset'],
                                      length=len(augmented_target_audio),
                                      audio_in=augmented_noise_audio)

    if record['snr'] < -96:
        # Special case for zeroing out target data
        record['target_snr_gain'] = 0
        record['noise_snr_gain'] = 1
        record['class_count'] = [[0] * len(inner) for inner in record['class_count']]
        # Setting target_gain to zero will cause the truth to be all zeros.
        record['target_gain'] = 0
    elif record['snr'] > 96:
        # Special case for zeroing out noise data
        record['target_snr_gain'] = 1
        record['noise_snr_gain'] = 0
    else:
        target_energy = np.mean(np.square(np.single(augmented_target_audio)))
        noise_energy = np.mean(np.square(np.single(noise_segment)))
        noise_gain = np.sqrt(target_energy / noise_energy) / 10 ** (
                record['snr'] / 20)

        # Check for noise_gain > 1 to avoid clipping
        if noise_gain > 1:
            record['target_snr_gain'] = 1 / noise_gain
            record['noise_snr_gain'] = 1
        else:
            record['target_snr_gain'] = 1
            record['noise_snr_gain'] = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = np.single(augmented_target_audio) * record['target_snr_gain']
    gain_adjusted_noise_audio = np.single(noise_segment) * record['noise_snr_gain']
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio

    if any(abs(mixture_audio) >= 32768):
        # Clipping occurred; lower gains to bring audio within int16 bounds
        gain_adjustment = 32760 / max(abs(mixture_audio))
        record['target_snr_gain'] *= gain_adjustment
        record['noise_snr_gain'] *= gain_adjustment

    return record



from typing import List
from typing import Union

from sonusai.mixture.get_mixtures_from_mixid import convert_mixid_to_list
from sonusai.mixture.mixture_data import get_feature_frames_in_mixture
from sonusai.mixture.segment import Segment


def get_file_frame_segments(mixdb: dict, mixid: Union[str, List[int]] = ':') -> dict:
    _mixid = convert_mixid_to_list(mixdb, mixid)
    file_frame_segments = dict()
    for m in _mixid:
        file_frame_segments[m] = Segment(mixdb['mixtures'][m]['o_frame_offset'],
                                         get_feature_frames_in_mixture(mixdb, m))
    return file_frame_segments



from glob import glob
from os import listdir
from os.path import dirname
from os.path import isabs
from os.path import isdir
from os.path import join
from os.path import splitext
from typing import List
from typing import Union

import sox

from sonusai import SonusAIError
from sonusai.mixture.config import raw_load_config
from sonusai.utils.expandvars import expandvars


def get_input_files(records: List[dict],
                    truth_settings: Union[List[dict], None] = None) -> List[dict]:
    if truth_settings is None:
        truth_settings = list()

    files = list()
    for record in records:
        append_input_files(files, record, truth_settings)
    return files


def append_input_files(files: List[dict],
                       in_record: Union[dict, str],
                       truth_settings: List[dict],
                       tokens: Union[dict, None] = None) -> None:
    if tokens is None:
        tokens = dict()

    if isinstance(in_record, dict):
        if 'target_name' in in_record:
            in_name = in_record['target_name']
        else:
            raise SonusAIError('Target list contained record without target_name')

        if 'truth_settings' in in_record:
            truth_settings = in_record['truth_settings']
    else:
        in_name = in_record

    in_name, new_tokens = expandvars(in_name)
    tokens.update(new_tokens)
    names = glob(in_name)
    if not names:
        raise SonusAIError(f'Could not find {in_name}. Make sure path exists')
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                append_input_files(files, child, truth_settings, tokens)
        else:
            try:
                if ext == '.txt':
                    with open(file=name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = expandvars(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                append_input_files(files, child, truth_settings, tokens)
                elif ext == '.yml':
                    try:
                        yml_config = raw_load_config(name)

                        if 'targets' in yml_config:
                            for record in yml_config['targets']:
                                append_input_files(files, record, truth_settings, tokens)
                    except Exception as e:
                        raise SonusAIError(f'Error processing {name}: {e}')
                else:
                    sox.file_info.validate_input_file(name)
                    duration = sox.file_info.duration(name)
                    for key, value in tokens.items():
                        name = name.replace(value, f'${key}')
                    entry = {
                        'name':     name,
                        'duration': duration,
                    }
                    if len(truth_settings) > 0:
                        entry['truth_settings'] = truth_settings
                        for truth_setting in entry['truth_settings']:
                            if 'function' in truth_setting and truth_setting['function'] == 'file':
                                truth_setting['config']['file'] = splitext(name)[0] + '.h5'
                    files.append(entry)
            except SonusAIError:
                raise
            except Exception as e:
                raise SonusAIError(f'Error processing {name}: {e}')



"""sonusai cbmixdb

usage: cbmixdb [-hv] (-d MIXDB) [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Input mixture database JSON file.
    -o OUTPUT, --output OUTPUT      Output class balanced mixture database JSON file.

Perform class balancing on a SonusAI mixture database.

Inputs:
    MIXDB   A SonusAI mixture database JSON file.

Outputs:
    OUTPUT  A SonusAI mixture database JSON file that has class balanced data.

    cbmixdb.log

"""
import json
import time
from random import seed

import numpy as np
import yaml
from docopt import docopt
from tqdm import tqdm

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import balance_classes
from sonusai.mixture import estimate_audio_length
from sonusai.mixture import get_feature_stats
from sonusai.mixture import get_total_class_count
from sonusai.mixture import load_mixdb
from sonusai.mixture import log_duration_and_sizes
from sonusai.mixture import process_mixture
from sonusai.mixture import process_target
from sonusai.mixture import read_all_audio
from sonusai.utils import p_tqdm_map
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


def cbmixdb(mixdb: dict,
            logging: bool = True,
            show_progress: bool = False) -> dict:
    seed(mixdb['seed'])

    if logging:
        logger.debug(f'Seed: {mixdb["seed"]}')

    targets = mixdb['targets']
    if logging:
        logger.debug('List of targets:')
        logger.debug(yaml.dump([sub['name'] for sub in targets], default_flow_style=False))

    noises = mixdb['noise_mix']['files']
    if logging:
        logger.debug('List of noises:')
        logger.debug(yaml.dump([sub['name'] for sub in noises], default_flow_style=False))

    target_augmentations = mixdb['target_augmentations']
    noise_augmentations = mixdb['noise_mix']['augmentations']

    if logging:
        logger.debug(f'SNRs: {mixdb["noise_mix"]["snrs"]}\n')
        logger.debug(f'Exhaustive noise: {mixdb["noise_mix"]["exhaustive"]}\n')

    (feature_ms,
     feature_samples,
     feature_step_ms,
     feature_step_samples,
     num_bands,
     stride) = get_feature_stats(feature_mode=mixdb['feature'],
                                 frame_size=mixdb['frame_size'],
                                 num_classes=mixdb['num_classes'],
                                 truth_mutex=mixdb['truth_mutex'])

    MP_DICT['mixdb'] = mixdb

    raw_noise_audio, raw_target_audio = read_all_audio(noises=noises, targets=targets, show_progress=show_progress)

    MP_DICT['raw_target_audio'] = raw_target_audio

    # Balance class data
    augmented_targets = balance_classes(mixdb=mixdb,
                                        raw_target_audio=raw_target_audio,
                                        show_progress=show_progress)

    noise_sets = len(noises) * len(noise_augmentations)
    target_sets = len(augmented_targets) * len(mixdb['noise_mix']['snrs'])
    total_mixtures = noise_sets * target_sets
    if logging:
        logger.info('')
        logger.info(f'Found {total_mixtures:,} mixtures to process')

    total_duration = 0
    for augmented_target in augmented_targets:
        length = int(targets[augmented_target['target_file_index']]['duration'] * sonusai.mixture.SAMPLE_RATE)
        augmentation = target_augmentations[augmented_target['target_augmentation_index']]
        if 'tempo' in augmentation:
            length /= augmentation['tempo']
        if length % feature_step_samples:
            length += feature_step_samples - int(length % feature_step_samples)
        total_duration += float(length) / sonusai.mixture.SAMPLE_RATE
    total_duration *= len(mixdb['noise_mix']['snrs'])
    total_duration *= noise_sets

    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb['num_classes'],
                               feature_step_samples=feature_step_samples,
                               num_bands=num_bands,
                               stride=stride,
                               desc='Estimated')
        logger.info(f'Feature shape:        {stride} x {num_bands} ({stride * num_bands} total params)')
        logger.info(f'Feature samples:      {feature_samples} samples ({feature_ms} ms)')
        logger.info(f'Feature step samples: {feature_step_samples} samples ({feature_step_ms} ms)')

    # Get indices and offsets
    mixtures = [[] for _ in range(noise_sets)]
    n_id = 0
    for noise_index, noise in enumerate(noises):
        for noise_augmentation_index, noise_augmentation in enumerate(noise_augmentations):
            mixtures[n_id] = [[] for _ in range(target_sets)]
            t_id = 0
            noise_offset = 0
            noise_length = estimate_audio_length(audio_in=raw_noise_audio[noise_index],
                                                 augmentation=noise_augmentation,
                                                 length_common_denominator=1)
            for target in augmented_targets:
                target_index = target['target_file_index']
                target_augmentation_index = target['target_augmentation_index']
                target_augmentation = mixdb['target_augmentations'][target_augmentation_index]

                for snr in mixdb['noise_mix']['snrs']:
                    mixtures[n_id][t_id] = {
                        'target_file_index':         target_index,
                        'noise_file_index':          noise_index,
                        'noise_offset':              noise_offset,
                        'target_augmentation_index': target_augmentation_index,
                        'noise_augmentation_index':  noise_augmentation_index,
                        'snr':                       snr,
                        'target_snr_gain':           np.NaN,
                        'noise_snr_gain':            np.NaN,
                        'samples':                   target['samples'],
                        'target_gain':               target['target_gain'],
                        'class_count':               target['class_count'],
                    }
                    t_id += 1

                    target_length = estimate_audio_length(audio_in=raw_target_audio[target_index],
                                                          augmentation=target_augmentation,
                                                          length_common_denominator=feature_step_samples)
                    noise_offset = int((noise_offset + target_length) % noise_length)

            n_id += 1

    # Fill in the details
    progress = tqdm(total=total_mixtures, desc='cbmixdb', disable=not show_progress)
    for n_id in range(len(mixtures)):
        MP_DICT['augmented_noise_audio'] = apply_augmentation(
            audio_in=raw_noise_audio[mixtures[n_id][0]['noise_file_index']],
            augmentation=noise_augmentations[mixtures[n_id][0]['noise_augmentation_index']],
            length_common_denominator=1)
        mixtures[n_id] = p_tqdm_map(_process_mixture, mixtures[n_id], progress=progress)
    progress.close()

    # Flatten mixtures
    mixdb['mixtures'] = [item for sublist in mixtures for item in sublist]
    mixdb['class_count'] = get_total_class_count(mixdb)

    total_samples = sum([sub['samples'] for sub in mixdb['mixtures']])
    total_duration = total_samples / sonusai.mixture.SAMPLE_RATE
    if logging:
        log_duration_and_sizes(total_duration=total_duration,
                               num_classes=mixdb['num_classes'],
                               feature_step_samples=feature_step_samples,
                               num_bands=num_bands,
                               stride=stride,
                               desc='Actual')

    return mixdb


def _process_target(record: dict) -> dict:
    return process_target(record=record, mixdb=MP_DICT['mixdb'], raw_target_audio=MP_DICT['raw_target_audio'])


def _process_mixture(record: dict) -> dict:
    return process_mixture(record=record,
                           mixdb=MP_DICT['mixdb'],
                           raw_target_audio=MP_DICT['raw_target_audio'],
                           augmented_noise_audio=MP_DICT['augmented_noise_audio'])


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']
        mixdb_name = args['--mixdb']
        output_name = args['--output']

        start_time = time.monotonic()

        log_name = 'cbmixdb.log'
        create_file_handler(log_name)
        update_console_handler(verbose)
        initial_log_messages('cbmixdb')

        logger.info(f'\nLoad mixture database from {mixdb_name}')
        mixdb = load_mixdb(mixdb_name)

        with open(file=output_name, mode='w') as file:
            json.dump(cbmixdb(mixdb=mixdb, show_progress=True), file, indent=2)
            logger.info(f'Wrote class balanced mixture database for {mixdb_name} to {output_name}')

        end_time = time.monotonic()
        logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



import numpy as np


def get_next_noise(offset_in: int, length: int, audio_in: np.ndarray) -> (np.ndarray, int):
    audio_out = np.take(audio_in, range(offset_in, offset_in + length), mode='wrap')
    offset_out = (offset_in + length) % len(audio_in)
    return audio_out, offset_out



import json

import h5py
import numpy as np


def read_feature_data(filename: str) -> (dict, np.ndarray, np.ndarray, np.ndarray):
    """Read mixdb, feature, truth_f, and segsnr data from given HDF5 file and return them as a tuple."""
    with h5py.File(name=filename, mode='r') as f:
        return json.loads(f.attrs['mixdb']), np.array(f['feature']), np.array(f['truth_f']), np.array(f['segsnr'])



import sonusai
from sonusai import logger
from sonusai.utils.human_readable_size import human_readable_size
from sonusai.utils.seconds_to_hms import seconds_to_hms


def log_duration_and_sizes(total_duration: int,
                           num_classes: int,
                           feature_step_samples: int,
                           num_bands: int,
                           stride: int,
                           desc: str) -> None:
    total_samples = total_duration * sonusai.mixture.SAMPLE_RATE
    mixture_bytes = total_samples * sonusai.mixture.BIT_DEPTH / 8
    truth_t_bytes = total_samples * num_classes * sonusai.mixture.FLOAT_BYTES
    feature_bytes = total_samples / feature_step_samples * stride * num_bands * sonusai.mixture.FLOAT_BYTES
    truth_f_bytes = total_samples / feature_step_samples * num_classes * sonusai.mixture.FLOAT_BYTES

    logger.info('')
    logger.info(f'{desc} duration:   {seconds_to_hms(seconds=total_duration)}')
    logger.info('{desc} sizes:')
    logger.info(f' mixture:             {human_readable_size(mixture_bytes, 1)}')
    logger.info(f' truth_t:             {human_readable_size(truth_t_bytes, 1)}')
    logger.info(f' feature:             {human_readable_size(feature_bytes, 1)}')
    logger.info(f' truth_f:             {human_readable_size(truth_f_bytes, 1)}')



"""sonusai mkwav

usage: mkwav [-hvn] FILE...

options:
   -h, --help
   -v, --verbose        Be verbose.
   -n, --dry-run        Don't actually create files, just show what will be done.

The mkwav command creates WAV files from audio data contained in genft .h5 files.
Specifically, it looks for these datasets: target, noise, and mixture.

"""
import wave
from os.path import splitext

import h5py
from docopt import docopt

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import mixture
from sonusai import update_console_handler
from sonusai.utils import trim_docstring


def mkwav(files: list, dry_run: bool = False, verbose: bool = False):
    update_console_handler(verbose)
    initial_log_messages('mkwav')

    audio_names = ['target', 'noise', 'mixture']

    if dry_run:
        logger.info('Dry run')

    for file in files:
        base_name = splitext(file)[0]
        if verbose:
            logger.info(f'Processing {file}')
        try:
            with h5py.File(file, 'r') as f:
                for audio_name in audio_names:
                    if audio_name in f:
                        out_name = base_name + '_' + audio_name + '.wav'
                        audio = f['/' + audio_name][:]
                        if verbose:
                            logger.info(f'Writing {len(audio)} samples to {out_name}')
                        if not dry_run:
                            with wave.open(out_name, mode='w') as w:
                                w.setnchannels(mixture.CHANNEL_COUNT)
                                w.setsampwidth(mixture.SAMPLE_BYTES)
                                w.setframerate(mixture.SAMPLE_RATE)
                                w.writeframesraw(audio)
        except Exception as e:
            logger.error(f'Error processing {file}: {e}')


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        log_name = 'mkwav.log'
        create_file_handler(log_name)

        files = args['FILE']
        if not isinstance(files, list):
            files = [files]

        mkwav(files=files, dry_run=args['--dry-run'], verbose=args['--verbose'])

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



from os.path import exists
from os.path import expandvars
from typing import List

import numpy as np
from tqdm import tqdm

from sonusai import SonusAIError
from sonusai.mixture.apply_augmentation import apply_augmentation
from sonusai.mixture.read_audio import read_audio
from sonusai.utils import p_tqdm_map

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


def build_noise_audio_db(mixdb: dict, show_progress: bool = False) -> List[dict]:
    """Build a list of noise audio data.

    Returns a list of dictionaries, each with the following keys:
        'file'            file index
        'augmentation'    augmentation index
        'audio'           audio samples
    """
    db = list()
    for file_index in tqdm(sorted(list(set([sub['noise_file_index'] for sub in mixdb['mixtures']]))),
                           desc='Read noise audio', disable=not show_progress):
        audio_in = read_audio(name=mixdb['noise_mix']['files'][file_index]['name'])
        for augmentation_index in sorted(list(set([sub['noise_augmentation_index'] for sub in mixdb['mixtures']]))):
            db.append({
                'file':         file_index,
                'augmentation': augmentation_index,
                'audio':        apply_augmentation(audio_in=audio_in,
                                                   augmentation=mixdb['noise_mix']['augmentations'][augmentation_index],
                                                   length_common_denominator=1)
            })
    return db


def get_noise_audio_from_db(db: List[dict], file_index: int, augmentation_index: int) -> np.ndarray:
    """Locate a record in a noise audio database using file and augmentation indices and return the audio samples."""
    for record in db:
        if record['file'] == file_index and record['augmentation'] == augmentation_index:
            return record['audio']

    raise SonusAIError(f'Could not find file_index {file_index} and augmentation_index '
                       f'{augmentation_index} in noise database')


def build_target_audio_db(mixdb: dict, show_progress: bool = False) -> list:
    """Build a list of target audio data.

    Returns a list of dictionaries, each with the following keys:
        'file'  file index
        'audio' audio samples
    """
    MP_DICT['mixdb'] = mixdb

    indices = sorted(list(set([sub['target_file_index'] for sub in mixdb['mixtures']])))
    progress = tqdm(total=len(indices), desc='Read target audio', disable=not show_progress)
    db = p_tqdm_map(_read_target_audio, indices, progress=progress)
    return db


def _read_target_audio(file_index: int) -> dict:
    """Parallel target audio reader kernel."""
    mixdb = MP_DICT['mixdb']
    return {
        'file':  file_index,
        'audio': read_audio(name=mixdb['targets'][file_index]['name'])
    }


def get_target_audio_from_db(db: List[dict], file_index: int) -> np.ndarray:
    """Locate a record in a target audio database using file index and return the audio samples."""
    for record in db:
        if record['file'] == file_index:
            return record['audio']

    raise SonusAIError(f'Could not find file_index {file_index} in target database')


def check_audio_files_exist(mixdb: dict) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist."""
    for file_index in sorted(list(set(sub['noise_file_index'] for sub in mixdb['mixtures']))):
        file_name = expandvars(mixdb['noise_mix']['files'][file_index]['name'])
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')

    for file_index in sorted(list(set(sub['target_file_index'] for sub in mixdb['mixtures']))):
        file_name = expandvars(mixdb['targets'][file_index]['name'])
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')


def read_all_audio(noises: List[dict],
                   targets: List[dict],
                   show_progress: bool = False) -> (List[np.ndarray], List[np.ndarray]):
    """Read in all audio data beforehand to avoid reading it multiple times in a loop."""
    names = [item['name'] for item in noises]
    progress = tqdm(total=len(names), desc='Read noise audio', disable=not show_progress)
    raw_noise_audio = p_tqdm_map(read_audio, names, progress=progress)
    progress.close()

    names = [item['name'] for item in targets]
    progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
    raw_target_audio = p_tqdm_map(read_audio, names, progress=progress)
    progress.close()

    return raw_noise_audio, raw_target_audio
